# © Copyright 2024 Hewlett Packard Enterprise Development LP
import argparse
import datetime
from typing import Any, List

import aiolirest
from aioli import cli
from aioli.cli import render
from aioli.cli.errors import CliError
from aioli.common.api import authentication
from aioli.common.declarative_argparse import Arg, Cmd, Group
from aiolirest.models.deployment_token import DeploymentToken
from aiolirest.models.deployment_token_patch_request import DeploymentTokenPatchRequest
from aiolirest.models.deployment_token_request import DeploymentTokenRequest


def format_token(t: DeploymentToken) -> List[Any]:
    result = [t.id, t.description, t.username, t.deployment, t.expiration, t.revoked]
    return result


@authentication.required
def create_token(args: argparse.Namespace) -> None:
    """Create a deployment token with the provided arguments.

    Invoke the Aioli Tokens API to create the deployment token. Print the ID of the deployment
    token created on the console.

    Args:
        args: command line arguments provided by the user. The deployment name argument
        is required, and the description and expiration arguments are optional.
    """
    with cli.setup_session(args) as session:
        tokens_api = aiolirest.TokensApi(session)

    # Check the format of the expiration provided by the user. If it doesn't include a timezone,
    # convert the expiration in to the ISO 8601 format with timezone details.
    if args.expiration:
        # Continue if the expiration is in ISO 8601 format and includes timezone.
        try:
            datetime.datetime.strptime(args.expiration, "%Y-%m-%dT%H:%M:%S%z")
        except Exception:
            # else if the expiration is in date and time format or date only format,
            # convert it to ISO 8601 format with timezone.
            accepted_formats = ["%Y-%m-%d %H:%M:%S", "%Y-%m-%d"]
            for datetime_format in accepted_formats:
                try:
                    # Adding astimezone method ensure we capture the user's timezone.
                    expiration_date = datetime.datetime.strptime(
                        args.expiration, datetime_format
                    ).astimezone()
                except Exception:
                    pass
                else:
                    args.expiration = expiration_date.isoformat()
                    break

    token_request = DeploymentTokenRequest(
        user=args.username,
        deployment=args.deployment,
        description=args.description,
        expiration=args.expiration,
    )
    token = tokens_api.tokens_post(request=token_request)
    print(token.id)


@authentication.required
def list_tokens(args: argparse.Namespace) -> None:
    """List active deployment tokens accessible to the current user.

    Invoke the Aioli Tokens API and fetch deployment tokens accessible to the current user.
    Only active tokens are listed by default. Adding the all flag will get all tokens for the
    all users if the user is has an admin role and all tokens for the current user otherwise.
    Format and display the deployment tokens on the console as a table by default. Output can be
    formatted as JSON or CSV based on the json or csv flag provided by the user.

    Args:
        args: command line arguments provided by the user. Includes json or csv flag to
        indicating output format. Includes all flag indicating token list should
        include all tokens for all users for admin users and all tokens for current user for other
        users.
    """
    with cli.setup_session(args) as session:
        tokens_api = aiolirest.TokensApi(session)
        if args.all:
            response = tokens_api.tokens_get(all="")
        else:
            response = tokens_api.tokens_get()

    if args.json:
        render.print_json([token.to_dict() for token in response])
    elif args.yaml:
        print(render.format_object_as_yaml([token.to_dict() for token in response]))
    else:
        headers = ["ID", "Description", "Username", "Deployment", "Expiration", "Revoked"]
        values = [format_token(t) for t in response]
        render.tabulate_or_csv(headers, values, args.csv)


@authentication.required
def get_token(args: argparse.Namespace) -> None:
    """Get the deployment token with the provided ID.

    Invoke the Aioli Tokens API to get the deployment token with the provided ID. Display the
    deployment token details in JSON format on the console.

    Args:
        args: command line arguments provided by the user. Includes ID for the deployment
        token.
    """
    with cli.setup_session(args) as session:
        tokens_api = aiolirest.TokensApi(session)
        token = tokens_api.tokens_id_get(args.id)
        if args.json:
            render.print_json(token.to_json())
        else:
            token_yaml = render.format_object_as_yaml(token.to_dict())
            print(token_yaml)


@authentication.required
def delete_token(args: argparse.Namespace) -> None:
    """Delete the deployment token with the provided ID.

    Invoke the Aioli Tokens API to delete the deployment token with the provided ID.

    Args:
        args: command line arguments provided by the user. Includes ID for the deployment
        token.
    """
    with cli.setup_session(args) as session:
        tokens_api = aiolirest.TokensApi(session)
        tokens_api.tokens_id_delete(args.id)


@authentication.required
def update_token(args: argparse.Namespace) -> None:
    """Update the description for the deployment token with the provided ID.

    Invoke the Aioli Tokens API to update the description for the deployment token with the
    provided ID.

    Args:
        args: command line arguments provided by the user. It contains deployment token ID
        and description.
    """
    if not args.description:
        raise CliError("No description provided. Use 'aioli token update -h' for usage.")
    with cli.setup_session(args) as session:
        tokens_api = aiolirest.TokensApi(session)
        token_patch_request = DeploymentTokenPatchRequest(
            description=args.description,
        )
        tokens_api.tokens_id_patch(args.id, token_patch_request)


@authentication.required
def revoke_token(args: argparse.Namespace) -> None:
    """Revoke the deployment token with the provided ID.

    Invoke the Aioli Tokens API to revoke the deployment token with the provided ID.

    Args:
        args: command line arguments provided by the user. Includes ID for the deployment
        token to be revoked.
    """
    with cli.setup_session(args) as session:
        tokens_api = aiolirest.TokensApi(session)
        token_patch_request = DeploymentTokenPatchRequest(
            revoked=True,
        )
        tokens_api.tokens_id_patch(args.id, token_patch_request)


args_description = [
    Cmd(
        "t|oken|s",
        None,
        "manage deployment tokens",
        [
            Cmd(
                "create",
                create_token,
                "create deployment token",
                [
                    Arg(
                        "deployment",
                        help="The deployment for which the token will be created.",
                    ),
                    Arg(
                        "--username",
                        help="The username for whom the token will be created. "
                        "If no username is provided, the current user's username will be used.",
                    ),
                    Arg(
                        "--description",
                        help="Description for the deployment token. "
                        "Enclose in quotes if the description contains spaces.",
                    ),
                    Arg("--expiration", help="Expiration date for the token."),
                ],
            ),
            Cmd(
                "list ls",
                list_tokens,
                "list the deployment tokens",
                [
                    Group(
                        Arg("--csv", action="store_true", help="print as CSV"),
                        Arg("--json", action="store_true", help="print as JSON"),
                        Arg("--yaml", action="store_true", help="print as YAML"),
                    ),
                    Arg(
                        "--all",
                        action="store_true",
                        help="Get all tokens for all users if you have an Admin role. Otherwise, "
                        "get all tokens accessible to you.",
                    ),
                ],
                is_default=True,
            ),
            Cmd(
                "show",
                get_token,
                "show the details of the deployment token with given ID",
                [
                    Arg("id", help="ID of the token"),
                    Group(
                        Arg("--json", action="store_true", help="print as JSON"),
                        Arg("--yaml", action="store_true", help="print as YAML"),
                    ),
                ],
            ),
            Cmd(
                "update",
                update_token,
                "update the description for deployment token with given ID",
                [
                    Arg("id", help="ID of the token"),
                    Arg(
                        "--description",
                        help="New description for the deployment token. "
                        "Enclose in quotes if the description contains spaces.",
                    ),
                ],
            ),
            Cmd(
                "revoke",
                revoke_token,
                "revoke the deployment token with given ID",
                [
                    Arg("id", help="ID of the token"),
                ],
            ),
            Cmd(
                "delete",
                delete_token,
                "delete the deployment token with given ID",
                [
                    Arg("id", help="ID of the token"),
                ],
            ),
        ],
    )
]  # type: List[Any]



# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
import datetime
import enum
import json
import math
import numbers
import uuid
from typing import Any, Optional, SupportsFloat, cast


def json_encode(obj: Any, indent: Optional[str] = None, sort_keys: bool = False) -> str:
    """
    Encode things as json, with an extra preprocessing step to handle some non-standard types.

    Note: json has a "default" argument that accepts something like our preprocessing step,
    except it is only invoked for non-native types (i.e. no catching nan or inf floats).
    """
    import numpy as np

    def jsonable(obj: Any) -> Any:
        if isinstance(obj, (str, bool, type(None))):
            # Needs no fancy encoding.
            return obj
        if isinstance(obj, numbers.Integral):
            # int, np.int64, etc.
            return int(obj)
        if isinstance(obj, numbers.Number):
            obj = cast(SupportsFloat, obj)
            # float, np.float64, etc.  Serialize nan/±infinity as strings.
            if math.isnan(obj):
                return "NaN"
            if math.isinf(obj):
                return "Infinity" if float(obj) > 0.0 else "-Infinity"
            return float(obj)
        if isinstance(obj, bytes):
            # Assume bytes are utf8 (json can't encode arbitrary binary data).
            return obj.decode("utf8")
        if isinstance(obj, (list, tuple)):
            # Recurse into lists.
            return [jsonable(v) for v in obj]
        if isinstance(obj, dict):
            # Recurse into dicts.
            return {k: jsonable(v) for k, v in obj.items()}
        if isinstance(obj, np.ndarray):
            # Expand arrays into lists, then recurse.
            return jsonable(obj.tolist())
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        if isinstance(obj, enum.Enum):
            return obj.name
        if isinstance(obj, uuid.UUID):
            return str(obj)
        # Objects that provide their own custom JSON serialization.
        if hasattr(obj, "__json__"):
            return obj.__json__()
        raise TypeError("Unserializable object {} of type {}".format(obj, type(obj)))

    return json.dumps(jsonable(obj), indent=indent, sort_keys=sort_keys)



# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
import argparse
from typing import Optional

import aiolirest
from aioli.common import api, util
from aioli.common.api import authentication


def setup_session_no_auth(host: str) -> aiolirest.ApiClient:
    host = util.prepend_protocol(host)

    return aiolirest.ApiClient(authentication.get_rest_config(host))


def setup_session(
    args: argparse.Namespace, controller: Optional[str] = None
) -> aiolirest.ApiClient:
    # Defining the host is optional and defaults to http://localhost:8080
    # See configuration.py for a list of all supported configuration parameters.
    if controller is not None:
        host = controller
    else:
        host = args.controller

    host = util.prepend_protocol(host)

    configuration = authentication.get_rest_config(host)

    token: Optional[str] = None

    if authentication.cli_auth is not None:
        token = authentication.cli_auth.get_session_token()
    if token is None:
        token = util.get_aioli_user_token_from_env()
        # Until something more permanent is available, import a token from the environment.
        if token is None:
            raise api.errors.NotFoundException("AIOLI_USER_TOKEN not defined")

    configuration.api_key["ApiKeyAuth"] = "Bearer " + token

    return aiolirest.ApiClient(configuration)



# © Copyright 2024 Hewlett Packard Enterprise Development LP
import atexit
import contextlib
import json
import logging
import os
import pathlib
import tempfile
from typing import Dict, Iterator, Optional, Union, cast

import certifi
import filelock
import urllib3

from aioli.common import api, util


class Cert:
    def __init__(
        self,
        cert_pem: Optional[str] = None,
        noverify: bool = False,
        name: Optional[str] = None,
    ) -> None:
        if cert_pem is not None and noverify:
            raise AssertionError("you cannot set cert_pem with noverify=True")

        if name == "":
            name = None
        self._name = name

        if noverify:
            self._bundle = False  # type: Union[None, str, bool]
        elif cert_pem is None:
            self._bundle = None
        else:
            # Don't use NamedTemporaryFile, since it would make the file inaccessible by path on
            # Windows after this.
            # (see https://docs.python.org/3/library/tempfile.html#tempfile.NamedTemporaryFile)
            fd, combined_path = tempfile.mkstemp(prefix="aioli-controller-cert-")
            atexit.register(os.unlink, combined_path)

            with open(fd, "wb") as out:
                with open(certifi.where(), "rb") as base_certs:
                    out.write(base_certs.read())
                out.write(b"\n")
                out.write(cert_pem.encode("utf8"))

            self._bundle = combined_path

    @property
    def bundle(self) -> Union[None, str, bool]:
        """
        The path to a file containing an SSL certificate to trust specifically for the controller,
        if any, or False to disable cert verification entirely. If set to a path, it should always
        be a temporary file that we own and can delete.
        """
        return self._bundle

    @property
    def name(self) -> Optional[str]:
        """
        The name we use to verify the controller certificate.
        """
        return self._name


cli_cert = None  # type: Optional[Cert]


class CertStore:
    """
    CertStore represents a persistent file-based record of certificates, each associated with a
    particular controller url.
    """

    def __init__(self, path: pathlib.Path) -> None:
        self.path = path

    def _load_store_file(self) -> Dict[str, str]:
        if not self.path.exists():
            return {}

        try:
            try:
                with self.path.open() as f:
                    store = json.load(f)
            except json.JSONDecodeError:
                raise api.errors.CorruptCertificateCacheException()

            # Store must be a dictionary.
            if not isinstance(store, dict):
                raise api.errors.CorruptCertificateCacheException()

            # All keys are url's, all values are pem-encoded certs.
            for k, v in store.items():
                if not isinstance(k, str):
                    raise api.errors.CorruptCertificateCacheException()
                if not isinstance(v, str):
                    raise api.errors.CorruptCertificateCacheException()

            return cast(Dict[str, str], store)

        except api.errors.CorruptCertificateCacheException:
            self.path.unlink()
            raise

    @contextlib.contextmanager
    def _persistent_store(self) -> Iterator[Dict["str", "str"]]:
        """
        Yields a store that can be modified, and the modified result will be written back to file.
        """
        self.path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
        # Decide on paths for a lock file and a temp files (during writing).
        temp = pathlib.Path(str(self.path) + ".temp")
        lock = str(self.path) + ".lock"

        with filelock.FileLock(lock):
            store = self._load_store_file()

            # No need for try/finally, because we don't update the file after failures.
            yield store

            with temp.open("w") as f:
                json.dump(store, f, indent=4, sort_keys=True)
            temp.replace(self.path)

    def get_cert(self, url: str) -> Optional[str]:
        """
        get_cert returns the contents of a cert (if any) that has been associated with the given
        url.
        """
        if not self.path.exists():
            return None
        # Technically this doesn't have to be modfiable, but it is unlikely to matter.
        with self._persistent_store() as store:
            return store.get(url)

    def set_cert(self, url: str, cert_pem: str) -> None:
        with self._persistent_store() as store:
            store[url] = cert_pem

    def delete_cert(self, url: str) -> None:
        with self._persistent_store() as store:
            if url in store:
                del store[url]


def maybe_shim_old_cert_store(
    old_path: pathlib.Path, new_path: pathlib.Path, controller_url: str
) -> None:
    """
    maybe_shim_old_cert_store will detect when an old v0 cert store is present and will shim it to
    a v1 cert store.
    """
    if not old_path.exists():
        return None

    # Only try to shim when ONLY the old path exists.
    if not new_path.exists():
        with old_path.open("r") as f:
            pem_content = f.read()
        store = {controller_url: pem_content}
        with new_path.open("w") as f:
            json.dump(store, f, indent=4, sort_keys=True)

    old_path.unlink()


def default_store() -> pathlib.Path:
    return util.get_config_path().joinpath("certs.json")


def get_cert_path() -> Union[None, str, bool]:
    # Let the environment variable for CERT_FILE override the default store.
    env_path = os.environ.get("AIOLI_CONTROLLER_CERT_FILE")
    if env_path:
        if os.path.exists(env_path):
            return env_path
        elif env_path.lower() == "noverify":
            # Turn off warnings to add CERT validation.
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            return False
        else:
            logging.warning(
                f"AIOLI_CONTROLLER_CERT_FILE={env_path} path not found; continuing without cert"
            )

    if cli_cert is not None:
        return cli_cert.bundle

    return None


def default_load(
    controller_url: str,
    explicit_path: Optional[str] = None,
    explicit_cert_name: Optional[str] = None,
    explicit_noverify: bool = False,
) -> Cert:
    """
    default_load takes as input the user-specified certificate-related configurations, reads
    environment variable configs, reads configs from the default store on the filesystem, and
    returns the resulting Cert object.

    Having all of the default loading logic in one function makes it easy to invoke the same logic
    in both the cli and the python sdk.
    """
    # Any explicit args causes us to ignore environment variables and defaults.
    if explicit_path or explicit_cert_name or explicit_noverify:
        if explicit_path:
            with open(explicit_path, "r") as f:
                cert_pem = f.read()  # type: Optional[str]
        else:
            cert_pem = None
        return Cert(cert_pem=cert_pem, noverify=explicit_noverify, name=explicit_cert_name)

    # Let any environment variable for CERT_FILE override the default store.
    env_path = os.environ.get("AIOLI_CONTROLLER_CERT_FILE")
    noverify = False
    cert_pem = None
    if env_path:
        if env_path.lower() == "noverify":
            # Turn off warnings to add CERT validation.
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            noverify = True
        elif os.path.exists(env_path):
            with open(env_path, "r") as f:
                cert_pem = f.read()
        else:
            logging.warning(
                f"AIOLI_CONTROLLER_CERT_FILE={env_path} path not found; continuing without cert"
            )
    else:
        # Otherwise, look in the default location for cert_pem.
        store_path = default_store()
        cert_store = CertStore(path=store_path)
        old_path = util.get_config_path().joinpath("server.crt")
        maybe_shim_old_cert_store(old_path, store_path, controller_url)
        cert_pem = cert_store.get_cert(controller_url)

    env_name = os.environ.get("AIOLI_CONTROLLER_CERT_NAME")
    if env_name == "":
        env_name = None

    return Cert(cert_pem=cert_pem, noverify=noverify, name=env_name)



# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
from aioli.common.api import errors
from aioli.common.api._util import PageOpts
from aioli.common.api.request import (
    WebSocket,
    browser_open,
    delete,
    do_request,
    get,
    make_url,
    parse_master_address,
    patch,
    post,
    put,
    ws,
)



# © Copyright 2024 Hewlett Packard Enterprise Development LP
from argparse import Namespace
from typing import Any, List

import aiolirest
from aioli import cli
from aioli.cli import render
from aioli.common.api import authentication
from aioli.common.api.errors import NotFoundException
from aioli.common.declarative_argparse import Arg, ArgsDescription, Cmd, Group
from aiolirest.models.auto_scaling_template import AutoScalingTemplate
from aiolirest.models.auto_scaling_template_request import AutoScalingTemplateRequest
from aiolirest.models.autoscaling import Autoscaling
from aiolirest.models.configuration_resources import ConfigurationResources
from aiolirest.models.resource_profile import ResourceProfile
from aiolirest.models.resources_template import ResourcesTemplate
from aiolirest.models.resources_template_request import ResourcesTemplateRequest


@authentication.required
def list_resource_templates(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)
        response = api_instance.templates_resources_get()

    def format_resource_template(rt: ResourcesTemplate) -> List[Any]:
        assert rt is not None
        assert rt.resources is not None
        assert rt.resources.requests is not None
        assert rt.resources.requests.cpu is not None
        assert rt.resources.requests.gpu is not None
        assert rt.resources.requests.memory is not None
        assert rt.resources.limits is not None
        assert rt.resources.limits.cpu is not None
        assert rt.resources.limits.gpu is not None
        assert rt.resources.limits.memory is not None

        result = [
            rt.name,
            rt.description,
            "Request: " + rt.resources.requests.cpu + ", Limit: " + rt.resources.limits.cpu,
            "Request: " + rt.resources.requests.memory + ", Limit: " + rt.resources.limits.memory,
            "Request: " + rt.resources.requests.gpu + ", Limit: " + rt.resources.limits.gpu,
            rt.resources.gpu_type,
        ]
        return result

    if args.json:
        render.print_json([r.to_dict() for r in response])
    elif args.yaml:
        print(render.format_object_as_yaml([r.to_dict() for r in response]))
    else:
        headers = ["Name", "Description", "CPU", "Memory", "GPU", "GPU Type"]
        values = [format_resource_template(rt) for rt in response]
        render.tabulate_or_csv(headers, values, args.csv)


def lookup_resource_template(name: str, api: aiolirest.TemplatesApi) -> ResourcesTemplate:
    for rt in api.templates_resources_get():
        if rt.name == name:
            return rt
    raise NotFoundException(f"resource template {name} not found")


@authentication.required
def show_resource_template(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)

    resource_template = lookup_resource_template(args.name, api_instance)

    rtd = resource_template.to_dict()
    if args.json:
        render.print_json(rtd)
    else:
        print(render.format_object_as_yaml(rtd))


@authentication.required
def create_resource_template(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)

        rt = ResourcesTemplateRequest(
            description=args.description,
            name=args.name,
            resources=ConfigurationResources(
                gpuType=args.gpu_type,
                limits=ResourceProfile(
                    cpu=args.limits_cpu, gpu=args.limits_gpu, memory=args.limits_memory
                ),
                requests=ResourceProfile(
                    cpu=args.requests_cpu, gpu=args.requests_gpu, memory=args.requests_memory
                ),
            ),
        )
        api_instance.templates_resources_post(rt)


@authentication.required
def update_resource_template(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)
        found = lookup_resource_template(args.templatename, api_instance)

        assert found is not None
        assert found.resources is not None
        assert found.resources.limits is not None
        assert found.resources.requests is not None

        request = ResourcesTemplateRequest(
            name=found.name,
            description=found.description,
            resources=ConfigurationResources(
                gpuType=found.resources.gpu_type,
                limits=ResourceProfile(
                    cpu=found.resources.limits.cpu,
                    gpu=found.resources.limits.gpu,
                    memory=found.resources.limits.memory,
                ),
                requests=ResourceProfile(
                    cpu=found.resources.requests.cpu,
                    gpu=found.resources.requests.gpu,
                    memory=found.resources.requests.memory,
                ),
            ),
        )

        assert request is not None
        assert request.resources is not None
        assert request.resources.limits is not None
        assert request.resources.requests is not None

        orig_name = found.name

        if args.name is not None:
            request.name = args.name

        if args.description is not None:
            request.description = args.description

        if args.gpu_type is not None:
            request.resources.gpu_type = args.gpu_type

        if args.limits_cpu is not None:
            request.resources.limits.cpu = args.limits_cpu

        if args.limits_memory is not None:
            request.resources.limits.memory = args.limits_memory

        if args.limits_gpu is not None:
            request.resources.limits.gpu = args.limits_gpu

        if args.requests_cpu is not None:
            request.resources.requests.cpu = args.requests_cpu

        if args.requests_memory is not None:
            request.resources.requests.memory = args.requests_memory

        if args.requests_gpu is not None:
            request.resources.requests.gpu = args.requests_gpu

        headers = {"Content-Type": "application/json"}
        assert orig_name is not None
        api_instance.templates_resources_name_put(orig_name, request, _headers=headers)


@authentication.required
def delete_resource_template(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)
        found = lookup_resource_template(args.name, api_instance)
        assert found.name is not None
        api_instance.templates_resources_name_delete(found.name)


@authentication.required
def list_autoscaling_templates(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)
        response = api_instance.templates_autoscaling_get()

    def format_autoscaling_template(at: AutoScalingTemplate) -> List[Any]:
        assert at is not None
        assert at.description is not None
        assert at.auto_scaling is not None

        result = [
            at.name,
            at.description,
            at.auto_scaling.min_replicas,
            at.auto_scaling.max_replicas,
            at.auto_scaling.metric,
            at.auto_scaling.target,
        ]
        return result

    if args.json:
        render.print_json([r.to_dict() for r in response])
    elif args.yaml:
        print(render.format_object_as_yaml([r.to_dict() for r in response]))
    else:
        headers = ["Name", "Description", "MinReplicas", "MaxReplicas", "Metric", "Target"]
        values = [format_autoscaling_template(at) for at in response]
        render.tabulate_or_csv(headers, values, args.csv)


def lookup_autoscaling_template(name: str, api: aiolirest.TemplatesApi) -> AutoScalingTemplate:
    for at in api.templates_autoscaling_get():
        if at.name == name:
            return at
    raise NotFoundException(f"autoscaling template {name} not found")


@authentication.required
def show_autoscaling_template(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)

    autoscaling_template = lookup_autoscaling_template(args.name, api_instance)

    atd = autoscaling_template.to_dict()
    if args.json:
        render.print_json(atd)
    else:
        print(render.format_object_as_yaml(atd))


@authentication.required
def create_autoscaling_template(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)

        at = AutoScalingTemplateRequest(
            autoScaling=Autoscaling(
                maxReplicas=args.autoscaling_max_replicas,
                metric=args.autoscaling_metric,
                minReplicas=args.autoscaling_min_replicas,
                target=args.autoscaling_target,
            ),
            description=args.description,
            name=args.name,
        )
        api_instance.templates_autoscaling_post(at)


@authentication.required
def update_autoscaling_template(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)
        found = lookup_autoscaling_template(args.templatename, api_instance)

        assert found is not None
        assert found.auto_scaling is not None

        request = AutoScalingTemplateRequest(
            autoScaling=Autoscaling(
                maxReplicas=found.auto_scaling.max_replicas,
                metric=found.auto_scaling.metric,
                minReplicas=found.auto_scaling.min_replicas,
                target=found.auto_scaling.target,
            ),
            description=found.description,
            name=found.name,
        )

        orig_name = found.name

        assert request is not None
        assert request.auto_scaling is not None

        if args.name is not None:
            request.name = args.name

        if args.description is not None:
            request.description = args.description

        if args.autoscaling_min_replicas is not None:
            request.auto_scaling.min_replicas = args.autoscaling_min_replicas

        if args.autoscaling_max_replicas is not None:
            request.auto_scaling.max_replicas = args.autoscaling_max_replicas

        if args.autoscaling_metric is not None:
            request.auto_scaling.metric = args.autoscaling_metric

        if args.autoscaling_target is not None:
            request.auto_scaling.target = args.autoscaling_target

        headers = {"Content-Type": "application/json"}
        assert orig_name is not None
        api_instance.templates_autoscaling_name_put(orig_name, request, _headers=headers)


@authentication.required
def delete_autoscaling_template(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.TemplatesApi(session)
        found = lookup_autoscaling_template(args.name, api_instance)
        assert found.name is not None
        api_instance.templates_autoscaling_name_delete(found.name)


common_resource_template_args: ArgsDescription = [
    Arg("--description", help="Description of the resource template."),
    Arg("--gpu-type", help="GPU type required"),
    Arg("--limits-cpu", help="CPU limit"),
    Arg("--limits-memory", help="Memory limit"),
    Arg("--limits-gpu", help="GPU limit"),
    Arg("--requests-cpu", help="CPU request"),
    Arg("--requests-memory", help="Memory request"),
    Arg("--requests-gpu", help="GPU request"),
]

main_cmd = Cmd(
    "template|s",
    None,
    "manage templates",
    [
        Cmd(
            "resource|s",
            None,
            "manage resource templates",
            [
                Cmd(
                    "list ls",
                    list_resource_templates,
                    "list resource templates",
                    [
                        Group(
                            Arg("--csv", action="store_true", help="print as CSV"),
                            Arg("--json", action="store_true", help="print as JSON"),
                            Arg("--yaml", action="store_true", help="print as YAML"),
                        ),
                    ],
                ),
                Cmd(
                    "show",
                    show_resource_template,
                    "show a resource template",
                    [
                        Arg("name", help="The name of the resource template."),
                        Group(
                            Arg("--yaml", action="store_true", help="print as YAML", default=True),
                            Arg("--json", action="store_true", help="print as JSON"),
                        ),
                    ],
                ),
                Cmd(
                    "create",
                    create_resource_template,
                    "create a resource template",
                    [
                        Arg(
                            "name",
                            help="The name of the resource resource template. Must begin "
                            + "with a letter, but may contain letters, numbers, "
                            + "underscore, and hyphen.",
                        ),
                    ]
                    + common_resource_template_args,
                ),
                Cmd(
                    "delete",
                    delete_resource_template,
                    "delete a resource template",
                    [Arg("name", help="The name of the resource template.")],
                ),
                Cmd(
                    "update",
                    update_resource_template,
                    "update a resource template",
                    [
                        Arg("templatename", help="The name of the resource template."),
                        Arg(
                            "--name",
                            help="The new name of the resource template. Must begin "
                            "with a letter, but may contain letters, numbers, and hyphen.",
                        ),
                    ]
                    + common_resource_template_args,
                ),
            ],
        ),
        Cmd(
            "autoscaling|s",
            None,
            "manage autoscaling templates",
            [
                Cmd(
                    "list ls",
                    list_autoscaling_templates,
                    "list autoscaling templates",
                    [
                        Group(
                            Arg("--csv", action="store_true", help="print as CSV"),
                            Arg("--json", action="store_true", help="print as JSON"),
                            Arg("--yaml", action="store_true", help="print as YAML"),
                        ),
                    ],
                ),
                Cmd(
                    "show",
                    show_autoscaling_template,
                    "show an autoscaling template",
                    [
                        Arg("name", help="The name of the autoscaling template."),
                        Group(
                            Arg("--yaml", action="store_true", help="print as YAML", default=True),
                            Arg("--json", action="store_true", help="print as JSON"),
                        ),
                    ],
                ),
                Cmd(
                    "create",
                    create_autoscaling_template,
                    "create an autoscaling template",
                    [
                        Arg(
                            "name",
                            help="The name of the autoscaling template. Must begin with a letter, "
                            + "but may contain letters, numbers, underscore, and hyphen.",
                        ),
                        Arg("--description", help="Description of the autoscaling template."),
                        Arg(
                            "--autoscaling-min-replicas",
                            help="Minimum number of replicas.",
                            required="true",
                            type=int,
                        ),
                        Arg(
                            "--autoscaling-max-replicas",
                            help="Maximum number of replicas.",
                            required="true",
                            type=int,
                        ),
                        Arg(
                            "--autoscaling-metric",
                            help="Metric name which controls autoscaling. Possible values include "
                            + "concurrency, rps, cpu, or memory.",
                            required="true",
                        ),
                        Arg(
                            "--autoscaling-target",
                            help="Metric target value.",
                            required="true",
                            type=int,
                        ),
                    ],
                ),
                Cmd(
                    "delete",
                    delete_autoscaling_template,
                    "delete an autoscaling template",
                    [Arg("name", help="The name of the autoscaling template.")],
                ),
                Cmd(
                    "update",
                    update_autoscaling_template,
                    "update an autoscaling template",
                    [
                        Arg("templatename", help="The name of the autoscaling template."),
                        Arg(
                            "--name",
                            help="The name of the autoscaling template. Must begin with a letter, "
                            + "but may contain letters, numbers, underscore, and hyphen.",
                        ),
                        Arg("--description", help="Description of the autoscaling template."),
                        Arg(
                            "--autoscaling-min-replicas",
                            help="Minimum number of replicas.",
                            type=int,
                        ),
                        Arg(
                            "--autoscaling-max-replicas",
                            help="Maximum number of replicas.",
                            type=int,
                        ),
                        Arg(
                            "--autoscaling-metric",
                            help="Metric name which controls autoscaling. "
                            + "Possible values include concurrency, rps, cpu, or memory.",
                        ),
                        Arg("--autoscaling-target", help="Metric target value.", type=int),
                    ],
                ),
            ],
        ),
    ],
)

args_description = [main_cmd]  # type: List[Any]



# © Copyright 2024 Hewlett Packard Enterprise Development LP
import sys
import webbrowser
from argparse import Namespace
from getpass import getpass
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Any, Callable, List
from urllib.parse import parse_qs, urlparse

from aioli.common import api
from aioli.common.api import authentication
from aioli.common.api.errors import NotFoundException
from aioli.common.declarative_argparse import Arg, Cmd

CLI_REDIRECT_PORT = 49176


def handle_token(master_url: str, token: str) -> None:
    tmp_auth = {"Cookie": "auth={token}".format(token=token)}
    me = api.get(master_url, "api/v1/users/me", headers=tmp_auth, authenticated=False).json()

    token_store = authentication.TokenStore(master_url)
    token_store.set_token(me["username"], token)
    token_store.set_active(me["username"])

    print("Authenticated as {}.".format(me["username"]))


def make_handler(master_url: str, close_cb: Callable[[int], None]) -> Any:
    class TokenAcceptHandler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            try:
                """Serve a GET request."""
                token = parse_qs(urlparse(self.path).query)["token"][0]
                handle_token(master_url, token)

                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(b"You can close this window now.")
                close_cb(0)
            except Exception as e:
                print("Error authenticating: {}.".format(e))
                close_cb(1)

        def log_message(self, format: Any, *args: List[Any]) -> None:  # noqa: A002
            # Silence server logging.
            return

    return TokenAcceptHandler


def sso(parsed_args: Namespace) -> None:
    master_info = api.get(parsed_args.controller, "info", authenticated=False).json()
    try:
        sso_providers = master_info["sso_providers"]
    except KeyError:
        raise NotFoundException("No SSO providers data")
    if not sso_providers:
        print("No SSO providers found.")
        return
    elif not parsed_args.provider:
        if len(sso_providers) > 1:
            print("Provider must be specified when multiple are available.")
            return
        matched_provider = sso_providers[0]
    else:
        matching_providers = [
            p for p in sso_providers if p["name"].lower() == parsed_args.provider.lower()
        ]
        if not matching_providers:
            ps = ", ".join(p["name"].lower() for p in sso_providers)
            print("Provider {} unsupported. (Providers found: {})".format(parsed_args.provider, ps))
            return
        elif len(matching_providers) > 1:
            print("Multiple SSO providers found with name {}.".format(parsed_args.provider))
            return
        matched_provider = matching_providers[0]

    sso_url = matched_provider["sso_url"] + "?relayState=cli"

    if not parsed_args.headless:
        if webbrowser.open(sso_url):
            print(
                "Your browser should open and prompt you to sign on;"
                " if it did not, please visit {}".format(sso_url)
            )
            print("Killing this process before signing on will cancel authentication.")
            with HTTPServer(
                ("localhost", CLI_REDIRECT_PORT),
                make_handler(parsed_args.controller, lambda code: sys.exit(code)),
            ) as httpd:
                return httpd.serve_forever()

        print("Failed to open Web Browser. Falling back to --headless CLI mode.")

    example_url = f"Example: 'http://localhost:{CLI_REDIRECT_PORT}/?token=v2.public.[long_str]'"

    print(
        f"Please open this URL in your browser: '{sso_url}'\n"
        "After authenticating, copy/paste the localhost URL "
        f"from your browser into the prompt.\n{example_url}"
    )
    token = None
    while not token:
        user_input_url = getpass(prompt="\n(hidden) localhost URL? ")
        try:
            token = parse_qs(urlparse(user_input_url).query)["token"][0]
            handle_token(parsed_args.controller, token)
        except (KeyError, IndexError):
            print(f"Could not extract token from localhost URL. {example_url}")


def list_providers(parsed_args: Namespace) -> None:
    master_info = api.get(parsed_args.controller, "info", authenticated=False).json()

    try:
        sso_providers = master_info["sso_providers"]
    except KeyError:
        raise NotFoundException("No SSO providers data")

    if len(sso_providers) == 0:
        print("No SSO providers found.")
        return

    print("Available providers: " + ", ".join(provider["name"] for provider in sso_providers) + ".")


# fmt: off

args_description = [
    Cmd("auth", None, "manage auth", [
        Cmd("login", sso, "sign on with an auth provider", [
            Arg("-p", "--provider", type=str,
                help="auth provider to use (not needed if the Aioli master only supports"
                " one provider)"),
            Arg("--headless", action="store_true", help="force headless cli auth")
        ]),
        Cmd("list-providers", list_providers, "lists the available auth providers", []),
    ])
]  # type: List[Any]

# fmt: on



# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
import functools
import os
import pathlib
import platform
import random
import warnings
from argparse import Namespace
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
    overload,
)

T = TypeVar("T")


def sizeof_fmt(val: Union[int, float]) -> str:
    val = float(val)
    for unit in ("", "K", "M", "G", "T", "P", "E", "Z"):
        if abs(val) < 1024.0:
            return "%3.1f%sB" % (val, unit)
        val /= 1024.0
    return "%.1f%sB" % (val, "Y")


def get_default_controller_address() -> str:
    return os.environ.get(
        "AIOLI_CONTROLLER", os.environ.get("AIOLI_CONTROLLER_ADDR", "localhost:8080")
    )


def get_aioli_username_from_env() -> Optional[str]:
    return os.environ.get("AIOLI_USER")


def get_aioli_user_token_from_env() -> Optional[str]:
    return os.environ.get("AIOLI_USER_TOKEN")


def get_aioli_password_from_env() -> Optional[str]:
    return os.environ.get("AIOLI_PASS")


def debug_mode() -> bool:
    return os.getenv("AIOLI_DEBUG", "").lower() in ("true", "1", "yes")


def preserve_random_state(fn: Callable) -> Callable:
    """A decorator to run a function with a fork of the random state."""

    @functools.wraps(fn)
    def wrapped(*arg: Any, **kwarg: Any) -> Any:
        state = random.getstate()
        try:
            return fn(*arg, **kwarg)
        finally:
            random.setstate(state)

    return wrapped


def get_config_path() -> pathlib.Path:
    if os.environ.get("AIOLI_DEBUG_CONFIG_PATH"):
        return pathlib.Path(os.environ["AIOLI_DEBUG_CONFIG_PATH"])

    system = platform.system()
    if "Linux" in system and "XDG_CONFIG_HOME" in os.environ:
        config_path = pathlib.Path(os.environ["XDG_CONFIG_HOME"])
    elif "Darwin" in system:
        config_path = pathlib.Path.home().joinpath("Library").joinpath("Application Support")
    elif "Windows" in system and "LOCALAPPDATA" in os.environ:
        config_path = pathlib.Path(os.environ["LOCALAPPDATA"])
    else:
        config_path = pathlib.Path.home().joinpath(".config")

    return config_path.joinpath("aioli")


U = TypeVar("U", bound=Callable[..., Any])


def deprecated(message: Optional[str] = None) -> Callable[[U], U]:
    def decorator(func: U) -> U:
        @functools.wraps(func)
        def wrapper_deprecated(*args: Any, **kwargs: Any) -> Any:
            warning_message = (
                f"{func.__name__} is deprecated and will be removed in a future version."
            )
            if message:
                warning_message += f" {message}."
            warnings.warn(warning_message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return cast(U, wrapper_deprecated)

    return decorator


def prepend_protocol(host: str) -> str:
    host = only_prepend_protocol(host)
    return f"{host}/api/v1"


def only_prepend_protocol(host: str) -> str:
    # If neither http nor https is specified, supply the default of http.
    if not (host.startswith("http://") or host.startswith("https://")):
        host = f"http://{host}"
    return host


@overload
def chunks(lst: str, chunk_size: int) -> Iterator[str]: ...


@overload
def chunks(lst: Sequence[T], chunk_size: int) -> Iterator[Sequence[T]]: ...


def chunks(
    lst: Union[str, Sequence[T]], chunk_size: int
) -> Union[Iterator[str], Iterator[Sequence[T]]]:
    """
    Collect data into fixed-length chunks or blocks.  Adapted from the
    itertools documentation recipes.

    e.g. chunks('ABCDEFG', 3) --> ABC DEF G
    """
    for i in range(0, len(lst), chunk_size):
        yield lst[i : i + chunk_size]


def launch_dashboard(args: Namespace, dashboard_uri: Optional[str]) -> None:
    import webbrowser

    url = f"{only_prepend_protocol(args.controller)}{dashboard_uri}"
    if not webbrowser.open(url):
        print(f"Failed to open a browser window. Manually open the dashboard using: {url}")


def construct_environment(args: Namespace) -> Dict[str, str]:
    environment: Dict[str, str] = {}
    if args.env is None:
        return environment

    for entry in args.env:
        if entry is None:
            continue
        # split to name & value
        the_split = entry.split("=", maxsplit=1)
        name: str = the_split[0]
        value: str = ""
        if len(the_split) > 1:
            value = the_split[1]
        environment[name] = value
    return environment


def construct_arguments(args: Namespace) -> List[str]:
    arguments: List[str] = []
    if args.arg is None:
        return arguments

    for entry in args.arg:
        if entry is None:
            continue
        arguments.append(entry.strip())
    return arguments


def strtobool(val: str) -> bool:
    """
    A port of the distutils.util.strtobool function, removed in python 3.12.

    The only difference in this function is that any non-falsy value which is a non-empty string is
    accepted as a true value.  That small difference gives us a small headstart on this todo:

    TODO(MLG-1520): we should instead treat any nonempty string as "true".
    """
    return bool(val and val.lower() not in ("n", "no", "f", "false", "off", "0"))



# © Copyright 2024 Hewlett Packard Enterprise Development LP
# The maximum size of a WebSocket message that can be sent or received
# by the Determined agent and trial-runner. The master uses a different limit,
# because it uses the uwsgi WebSocket implementation; see
# `websocket-max-size` in `uwsgi.ini`.
MAX_WEBSOCKET_MSG_SIZE = 128 * 1024 * 1024

# The maximum HTTP request size that will be accepted by the master. This
# is intended as a safeguard to quickly drop overly large HTTP requests.
MAX_HTTP_REQUEST_SIZE = 128 * 1024 * 1024



# © Copyright 2024 Hewlett Packard Enterprise Development LP
"""
Functions for runtime correctness checks.

Please do not use these functions in the new code. Instead, use assert
with a second argument that explains what was asserted.
See https://github.com/determined-ai/determined/pull/4424#discussion_r991583690

The module contains similar functionality under two different naming schemes.
The first is for code that imports function symbols directly, e.g.,

  from check import check_true, check_false
  ...
  check_true(...)


The second scheme is intended for code that only imports the check module
itself.

  import check
  ...
  check.true(...)

The second scheme is generally preferred because it more clearly indicates a
module dependency at the call site and it reduces diff churn when checks are
added or removed.
"""
from typing import Any, Container, Optional, Sized, Tuple, Type, Union


class CheckFailedError(Exception):
    pass


def true(val: bool, reason: Optional[str] = None) -> None:
    if val:
        return

    msg = "CHECK FAILED! Got {}, expected True".format(val)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_true(val: bool, reason: Optional[str] = None) -> None:
    return true(val, reason)


def false(val: bool, reason: Optional[str] = None) -> None:
    if not val:
        return

    msg = "CHECK FAILED! Got {}, expected False".format(val)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_false(val: bool, reason: Optional[str] = None) -> None:
    return false(val, reason)


def is_none(val: Optional[Any], reason: Optional[str] = None) -> None:
    if val is None:
        return

    msg = "CHECK FAILED! Got {}, expected None".format(val)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_none(val: Optional[Any], reason: Optional[str] = None) -> None:
    return is_none(val, reason)


def is_not_none(val: Optional[Any], reason: Optional[str] = None) -> None:
    if val is not None:
        return

    msg = "CHECK FAILED! Got {}, expected non-None".format(val)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_not_none(val: Optional[Any], reason: Optional[str] = None) -> None:
    return is_not_none(val, reason)


def eq(val: Any, expected: Any, reason: Optional[str] = None) -> None:
    if val == expected:
        return

    msg = "CHECK FAILED! Got {}, expected {}".format(val, expected)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_eq(val: Any, expected: Any, reason: Optional[str] = None) -> None:
    return eq(val, expected, reason)


def not_eq(x: Any, y: Any, reason: Optional[str] = None) -> None:
    if x != y:
        return

    msg = "CHECK FAILED! Got {}, expected value != {}".format(x, y)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_not_eq(x: Any, y: Any, reason: Optional[str] = None) -> None:
    return not_eq(x, y, reason)


def gt(x: Any, y: Any, reason: Optional[str] = None) -> None:
    if x > y:
        return

    msg = "CHECK FAILED! Got {}, expected value > {}".format(x, y)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_gt(x: Any, y: Any, reason: Optional[str] = None) -> None:
    return gt(x, y, reason)


def gt_eq(x: Any, y: Any, reason: Optional[str] = None) -> None:
    if x >= y:
        return

    msg = "CHECK FAILED! Got {}, expected value >= {}".format(x, y)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_gt_eq(x: Any, y: Any, reason: Optional[str] = None) -> None:
    return gt_eq(x, y, reason)


def lt(x: Any, y: Any, reason: Optional[str] = None) -> None:
    if x < y:
        return

    msg = "CHECK FAILED! Got {}, expected value < {}".format(x, y)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_lt(x: Any, y: Any, reason: Optional[str] = None) -> None:
    return lt(x, y, reason)


def lt_eq(x: Any, y: Any, reason: Optional[str] = None) -> None:
    if x <= y:
        return

    msg = "CHECK FAILED! Got {}, expected value <= {}".format(x, y)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_lt_eq(x: Any, y: Any, reason: Optional[str] = None) -> None:
    return lt_eq(x, y, reason)


def equal_lengths(x: Sized, y: Sized, reason: Optional[str] = None) -> None:
    if len(x) == len(y):
        return

    msg = "CHECK FAILED! Expected lengths {} and {} to be equal".format(len(x), len(y))
    msg += "; values: {}, {}".format(x, y)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_eq_len(x: Sized, y: Sized, reason: Optional[str] = None) -> None:
    return equal_lengths(x, y, reason)


def len_eq(val: Sized, expected_len: int, reason: Optional[str] = None) -> None:
    val_len = len(val)
    if val_len == expected_len:
        return

    msg = "CHECK FAILED! Got length {}, expected length {}".format(val_len, expected_len)
    if reason is not None:
        msg += ": {}".format(reason)
    msg += ". Values: {}".format(val)

    raise CheckFailedError(msg)


def check_len(val: Sized, expected_len: int, reason: Optional[str] = None) -> None:
    return len_eq(val, expected_len, reason)


def is_in(val: Any, expected: Container[Any], reason: Optional[str] = None) -> None:
    if val in expected:
        return

    msg = "CHECK FAILED! "

    # Report a more concise error message when `expected` is a dict.
    if isinstance(expected, dict):
        msg += "'{}' is not in {}".format(val, list(expected.keys()))
    else:
        msg += "'{}' is not in {}".format(val, expected)

    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_in(val: Any, expected: Container[Any], reason: Optional[str] = None) -> None:
    return is_in(val, expected, reason)


def not_in(val: Any, expected: Container[Any], reason: Optional[str] = None) -> None:
    if val not in expected:
        return

    msg = "CHECK FAILED! Got {}, expected value not in {}".format(val, expected)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_not_in(val: Any, expected: Container[Any], reason: Optional[str] = None) -> None:
    return not_in(val, expected, reason)


def is_type(val: Any, expected: type, reason: Optional[str] = None) -> None:
    if type(val) == expected:
        return

    msg = "CHECK FAILED! {} has type {}, expected type {}".format(val, type(val), expected)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_type(val: Any, expected: type, reason: Optional[str] = None) -> None:
    return is_type(val, expected, reason)


def is_instance(
    val: Any, expected: Union[Type, Tuple[Type, ...]], reason: Optional[str] = None
) -> None:
    if isinstance(val, expected):
        return

    msg = "CHECK FAILED! {} has type {}, expected isinstance of {}".format(val, type(val), expected)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_isinstance(
    val: Any, expected: Union[Type, Tuple[Type, ...]], reason: Optional[str] = None
) -> None:
    return is_instance(val, expected, reason)


def is_not_instance(
    val: Any, expected: Union[Type, Tuple[Type, ...]], reason: Optional[str] = None
) -> None:
    if not isinstance(val, expected):
        return

    msg = "CHECK FAILED! {} has type {}, expected not isinstance of {}".format(
        val, type(val), expected
    )
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_not_isinstance(
    val: Any, expected: Union[Type, Tuple[Type, ...]], reason: Optional[str] = None
) -> None:
    return is_not_instance(val, expected, reason)


def is_subclass(val: Any, expected: type, reason: Optional[str] = None) -> None:
    if issubclass(val, expected):
        return

    msg = "CHECK FAILED! {} has type {}, expected issubclass of {}".format(val, type(val), expected)
    if reason is not None:
        msg += ": {}".format(reason)

    raise CheckFailedError(msg)


def check_issubclass(val: Any, expected: type, reason: Optional[str] = None) -> None:
    return is_subclass(val, expected, reason)



# © Copyright 2024 Hewlett Packard Enterprise Development LP
"""
A drop-in replacement for requests.request() which supports server name overriding.
"""
from typing import Any, Optional

import requests
import urllib3


class HTTPAdapter(requests.adapters.HTTPAdapter):
    """A new HTTPAdapter which honors the ServerName as a value for the verify arg."""

    def __init__(self, server_hostname: Optional[str], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.server_hostname = server_hostname

    def cert_verify(self, conn: Any, url: Any, verify: Any, cert: Any) -> None:
        super().cert_verify(conn, url, verify, cert)  # type: ignore
        if self.server_hostname is not None:
            # Set the server_hostname value of the urllib3 connection.
            conn.assert_hostname = self.server_hostname


class Session(requests.sessions.Session):
    def __init__(
        self, server_hostname: Optional[str], max_retries: Optional[urllib3.util.retry.Retry]
    ) -> None:
        super().__init__()
        if max_retries is None:
            # Override the https adapter.
            self.mount("https://", HTTPAdapter(server_hostname))
        else:
            self.mount("https://", HTTPAdapter(server_hostname, max_retries=max_retries))
            self.mount("http://", requests.adapters.HTTPAdapter(max_retries=max_retries))


def request(method: str, url: str, **kwargs: Any) -> requests.Response:
    server_hostname = kwargs.pop("server_hostname", None)
    max_retries = kwargs.pop("max_retries", None)
    with Session(server_hostname, max_retries) as session:
        out = session.request(method=method, url=url, **kwargs)  # type: requests.Response
        return out



# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
from typing import Any

import requests


class BadRequestException(Exception):
    def __init__(self, message: str, *args: Any) -> None:
        super().__init__(message, *args)
        self.message = message

    def __str__(self) -> str:
        return self.message


class BadResponseException(Exception):
    def __init__(self, message: str, *args: Any) -> None:
        super().__init__(message, *args)
        self.message = message

    def __str__(self) -> str:
        return self.message


class MasterNotFoundException(BadRequestException):
    pass


class APIException(BadRequestException):
    """Raised when an API operation has failed.

    APIException is a catchall for errors passed on from the REST API server that aren't
    otherwise classified as other types of exceptions.

    Attributes:
        response_error: A dict parsed from the Response body containing structured error information
        status_code: The HTTP status code of the response
        message: A string containing a human-friendly error message. Inherited from
                 BadRequestException
    """

    def __init__(self, response: requests.Response, *args: Any) -> None:
        """Initialization from the Response that it was raised from.

        Args:
            response: A requests.Response with a body that looks like:
                {
                    error: {
                        code: <int>,
                        reason: <error type>,
                        error: <detailed error message>
                    }
                }
        """
        try:
            self.response_error = response.json()["error"]
            m = self.response_error["error"]
        except (ValueError, KeyError):
            self.response_error = None
            m = response.text
        super().__init__(m, response, *args)
        self.status_code = response.status_code


class NotFoundException(APIException):
    def __init__(self, error_message: str) -> None:
        self.message = error_message
        self.status_code = 404


class ForbiddenException(BadRequestException):
    def __init__(self, username: str, message: str = ""):
        err_message = f"Forbidden({message})"
        if not (message == "invalid credentials" or message == "user not found"):
            err_message += ": Please contact your administrator in order to access this resource."

        super().__init__(message=err_message)
        self.username = username


class UnauthenticatedException(BadRequestException):
    def __init__(self, username: str):
        super().__init__(message="Unauthenticated: use 'aioli user login <username>' to login.")
        self.username = username


class CorruptTokenCacheException(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Attempted to read a corrupted token cache.  The store has been deleted; "
            "please try again.\n"
        )


class CorruptCertificateCacheException(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Attempted to read a corrupted certificate cache.  The store has been deleted; "
            "please try again.\n"
        )


class EmptyResultException(BadResponseException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class VersionRequiredException(BadResponseException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class DeleteFailedException(APIException):
    def __init__(self, error_message: str) -> None:
        self.message = error_message
        self.status_code = 200



# © Copyright 2024 Hewlett Packard Enterprise Development LP
import csv
import json
import os
import subprocess

import yaml


class TestCli:
    testdir = os.path.dirname(os.path.realpath(__file__))

    def test_version(self, setup_login: None) -> None:
        version_file = self.testdir + "/../../../../VERSION"
        expected = None
        with open(version_file, "r") as file:
            expected = file.read().rstrip()
            expected = "aioli " + expected
        assert 0 == os.system("aioli --version")
        actual = os.popen("aioli --version").read().strip()
        assert actual == expected

    def test_version_bad(self, setup_login: None) -> None:
        bad_version = "aioli 0.0.1-bad"
        assert 0 == os.system("aioli --version")
        actual = os.popen("aioli --version").read().strip()
        assert actual != bad_version

    def test_project_list(self, setup_login: None) -> None:
        # Validating table headers, expect to see the default project
        expected = (
            " Name    | Description         | Owner\n"
            "---------+---------------------+---------\n"
            " default | The default project | admin\n"
        )
        actual = subprocess.check_output(["aioli", "project", "list"]).decode("utf-8")
        assert actual == expected

    def test_project_create(self, setup_login: None) -> None:
        # Create a project and test for expected values
        assert (
            os.system(
                "aioli project create proj-2 " "--description 'the test project' " "--owner admin"
            )
            == 0
        )
        expected = "proj-2  | the test project    | admin"
        actual = subprocess.check_output(["aioli", "project", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

    def test_project_list_csv(self, setup_login: None) -> None:
        csv_output = subprocess.check_output(["aioli", "project", "list", "--csv"]).decode("utf-8")
        temp_file: str = "/tmp/test-cli-p.csv"
        with open(temp_file, "w") as file:
            file.write(csv_output)
        with open(temp_file, newline="") as file:
            reader = csv.DictReader(file)
            row1 = reader.__next__()
            assert row1["Name"] == "default", "Expected default"
            assert row1["Description"] == "The default project", "Expected The default project"
            assert row1["Owner"] == "admin", "Expected admin"
            row2 = reader.__next__()
            assert row2["Name"] == "proj-2", "Expected proj-2"
            assert row2["Description"] == "the test project", "Expected the test project"
            assert row2["Owner"] == "admin", "Expected admin"

    # fmt: off
    def test_project_list_json_yaml(self, setup_login: None) -> None:
        expected = (
            '[\n'       # noqa: Q000
            '  {\n'     # noqa: Q000
            '    "description": "the test project",\n'
            '    "name": "proj-2",\n'
            '    "owner": "admin"\n'
            '  },\n'    # noqa: Q000
            '  {\n'     # noqa: Q000
            '    "description": "The default project",\n'
            '    "name": "default",\n'
            '    "owner": "admin"\n'
            '  }\n'     # noqa: Q000
            ']\n'       # noqa: Q000
        )
        actual = subprocess.check_output(
            ["aioli", "project", "list", "--json"]).decode("utf-8")
        assert actual == expected

        expected_obj = json.loads(expected)
        actual_obj = yaml.safe_load(
            subprocess.check_output(["aioli", "project", "list", "--yaml"]).decode("utf-8"))

        assert actual_obj == expected_obj

    # fmt: on

    def test_project_update(self, setup_login: None) -> None:
        # Update project and test for expected values
        assert (
            os.system(
                "aioli project update proj-2 --name proj-3 "
                "--description 'the updated test project' "
                "--owner admin"
            )
            == 0
        )
        actual = subprocess.check_output(["aioli", "project", "list"]).decode("utf-8")
        expected = "proj-3  | the updated test project | admin"
        assert (actual.find(expected)) > 0

    def test_project_delete(self, setup_login: None) -> None:
        # Delete project and test for expected values
        assert os.system("aioli project delete proj-3") == 0
        actual = subprocess.check_output(["aioli", "project", "list"]).decode("utf-8")
        expected = "proj-3"
        assert (actual.find(expected)) < 0  # Expect to be not found

    def test_project_show(self, setup_login: None) -> None:
        # Show project and test for expected values
        expected = "description: The default project\nname: default\nowner: admin"
        result = subprocess.check_output(["aioli", "project", "show", "default"]).decode("utf-8")
        result_list = result.split("\n")
        filtered_result = ""

        # Filter out modifiedAt and id fields as they vary for each test run
        for line in result_list:
            if "modifiedAt" in line:
                continue
            if "id" in line:
                continue
            filtered_result = filtered_result + "\n" + line

        actual = filtered_result.strip()
        assert actual == expected

    def test_registry_list(self, setup_login: None) -> None:
        # Validating table headers, there are no entries yet
        assert os.system("aioli registry list") == 0
        expected = (
            " Name   | Type   | Access Key   | Bucket   | Secret Key   | Endpoint URL"
            "\n--------+--------+--------------+----------+--------------+----------------\n"
        )
        actual = subprocess.check_output(["aioli", "registry", "list"]).decode("utf-8")
        assert actual == expected

    def test_registry_create_and_update_insecure(self, setup_login: None) -> None:
        # Create a registry with insecure-https and test for expected values
        assert (
            os.system(
                "aioli registry create --type s3 --access-key minioadmin "
                "--secret-key minioadmin --bucket demo-bento-registry "
                "--endpoint-url http://10.30.89.14:30008 --insecure-https bento-registry-insecure"
            )
            == 0
        )

        # Newly created registry entry without id/modified as dict
        expected = yaml.safe_load(
            "accessKey: minioadmin\nbucket: demo-bento-registry\nendpointUrl: "
            "http://10.30.89.14:30008\ninsecureHttps: true\n"
            "name: bento-registry-insecure\nsecretKey: '**********'\ntype: s3\n\n"
        )
        actualstr = subprocess.check_output(
            ["aioli", "registry", "show", "bento-registry-insecure"]
        ).decode("utf-8")
        actual = yaml.safe_load(actualstr)
        del actual["id"]
        del actual["modifiedAt"]
        assert actual == expected

        # Update to secure
        if False:
            # This is blocked until CI is at python 3.9 -- then update
            # registry.py to use action=argparse.BoolOptionalAction
            assert (
                os.system("aioli registry update --no-insecure-https bento-registry-insecure") == 0
            )
            expected["insecureHttps"] = False
            actual = subprocess.check_output(
                ["aioli", "registry", "show", "bento-registry-insecure"]
            ).decode("utf-8")
            actual = yaml.safe_load(actual)
            del actual["id"]
            del actual["modifiedAt"]
            assert actual == expected

        # Cleanup test registry
        assert os.system("aioli registry delete bento-registry-insecure") == 0

    def test_registry_create(self, setup_login: None) -> None:
        # Create a registry entry and test for success
        assert (
            os.system(
                "aioli registry create --type s3 --access-key minioadmin "
                "--secret-key minioadmin --bucket demo-bento-registry "
                "--endpoint-url http://10.30.89.14:30008/ bento-registry"
            )
            == 0
        )

        # The following test is expected to fail since we already created the
        # registry and we are trying to recreate the same.
        expected = (
            "Failed to create a registry: model registry named 'bento-registry' exists already."
        )
        try:
            subprocess.check_output(
                [
                    "aioli",
                    "registry",
                    "create",
                    "--type",
                    "s3",
                    "--access-key",
                    "minioadmin",
                    "--secret-key",
                    "minioadmin",
                    "--bucket",
                    "demo-bento-registry",
                    "--endpoint-url",
                    "http://10.30.89.14:30008/",
                    "bento-registry",
                ],
                stderr=subprocess.STDOUT,
            ).decode("utf-8")
        except subprocess.CalledProcessError as e:
            actual = (e.output).decode("utf-8")
            assert (actual.find(expected)) >= 0

        # Check enforcement of --type values
        proc = subprocess.run(
            [
                "aioli",
                "registry",
                "create",
                "--type",
                "bob",
                "--access-key",
                "minioadmin",
                "--secret-key",
                "minioadmin",
                "--bucket",
                "demo-bento-registry",
                "--endpoint-url",
                "http://10.30.89.14:30008/",
                "bento-registry-wrong-type",
            ],
            capture_output=True,
        )
        assert proc.returncode == 1
        expected = (
            "Failed to create a registry: registry type must be one of the values "
            + "(s3, http, openllm, ngc, pfs, huggingface), provided 'bob'"
        )
        assert proc.stderr.decode("utf-8").find(expected) == 0

        # List the newly created registry entry and test for expected values
        expected = (
            "bento-registry | s3     | minioadmin   | demo-bento-registry | "
            "**********   | http://10.30.89.14:30008/"
        )
        actual = subprocess.check_output(["aioli", "registry", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

    def test_registry_model_list(self, setup_login: None) -> None:
        # List the models for the registry and test for expected values
        expected = (
            "Failed to lists the available models in a registry (for certain registry types): "
            "a registry of type s3 is unable to provide a list of supported models. "
            "Supported registries: (openllm, ngc, huggingface, pfs)"
        )
        outcome = subprocess.run(
            ["aioli", "registry", "models", "bento-registry"], capture_output=True
        )

        assert outcome.returncode == 1
        assert outcome.stderr.decode("utf-8").find(expected) == 0

    def test_registry_list_csv(self, setup_login: None) -> None:
        csv_output = subprocess.check_output(["aioli", "registry", "list", "--csv"]).decode("utf-8")
        temp_file: str = "/tmp/test-cli-r.csv"
        with open(temp_file, "w") as file:
            file.write(csv_output)
        with open(temp_file, newline="") as file:
            reader = csv.DictReader(file)
            for row in reader:
                assert row["Name"] == "bento-registry", "Expected bento-registry"
                assert row["Type"] == "s3", "Expected s3"
                assert row["Access Key"] == "minioadmin", "Expected minioadmin"
                assert row["Bucket"] == "demo-bento-registry", "Expected demo-bento-registry"
                assert row["Secret Key"] == "**********", "Expected **********"
                assert (
                    row["Endpoint URL"] == "http://10.30.89.14:30008/"
                ), "Expected http://10.30.89.14:30008/"

    # fmt: off
    def test_registry_list_json_yaml(self, setup_login: None) -> None:
        expected = (
            '[\n'    # noqa: Q000
            '  {\n'  # noqa: Q000
            '    "accessKey": "minioadmin",\n'
            '    "bucket": "demo-bento-registry",\n'
            '    "endpointUrl": "http://10.30.89.14:30008/",\n'
            '    "insecureHttps": false,\n'
            '    "name": "bento-registry",\n'
            '    "secretKey": "**********",\n'
            '    "type": "s3"\n'
            '  }\n'  # noqa: Q000
            ']\n'    # noqa: Q000
        )
        actual = subprocess.check_output(["aioli", "registry", "list", "--json"]).decode("utf-8")
        assert actual == expected

        expected_obj = json.loads(expected)
        actual_obj = yaml.safe_load(
            subprocess.check_output(["aioli", "registry", "list", "--yaml"]).decode("utf-8"))

        assert actual_obj == expected_obj

    # fmt: on

    def test_registry_update(self, setup_login: None) -> None:
        # Update registry and test for expected values
        subprocess.check_output(
            [
                "aioli",
                "registry",
                "update",
                "--type",
                "s3",
                "--access-key",
                "minioadmin",
                "--secret-key",
                "minioadmin",
                "--bucket",
                "demo-bento-registry",
                "--endpoint-url",
                "http://10.30.89.14:30008/",
                "--name",
                "bento-registry1",
                "bento-registry",
            ],
            stderr=subprocess.STDOUT,
        ).decode("utf-8")
        expected = (
            "bento-registry1 | s3     | minioadmin   | demo-bento-registry | "
            "**********   | http://10.30.89.14:30008/"
        )
        actual = subprocess.check_output(["aioli", "registry", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

        # Check enforcement of --type values
        proc = subprocess.run(
            [
                "aioli",
                "registry",
                "update",
                "--type",
                "bob",
                "bento-registry1",
            ],
            capture_output=True,
        )
        assert proc.returncode == 1
        expected = (
            "Failed to modify a registry: registry type must be one of the values "
            + "(s3, http, openllm, ngc, pfs, huggingface), provided 'bob'"
        )
        assert proc.stderr.decode("utf-8").find(expected) == 0

    def test_model_list(self, setup_login: None) -> None:
        # Test header row of model table, there are no entries at this point.
        assert os.system("aioli model list") == 0
        expected = (
            " Name   | Description   | Version   | URI   | Image   | Registry"
            "\n--------+---------------+-----------+-------+---------+------------\n"
        )
        actual = subprocess.check_output(["aioli", "model", "list"]).decode("utf-8")
        assert actual == expected

        # Test that --name and --all are mutually exclusive parameters
        assert os.system("aioli model list --all --name doesnt-matter") == 512

        # Create version 1 of a model
        command = [
            "aioli",
            "model",
            "create",
            "my-model",
            "--image",
            "fictional.registry.example/imagename",
            "--description",
            "the model description",
            "--requests-cpu",
            "1.0",
            "--requests-memory",
            "1Gi",
            "--requests-gpu",
            "1.0",
            "--limits-cpu",
            "2.5",
            "--limits-memory",
            "1Gi",
            "--limits-gpu",
            "1.0",
        ]

        # Run the command and capture stdout and stderr
        result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        # Print the stdout and stderr
        print("STDOUT:", result.stdout)
        print("STDERR:", result.stderr)

        # Check the exit status
        assert result.returncode == 0, f"Command failed with exit code {result.returncode}"

        # Create version 2
        assert os.system("aioli model update my-model " "--limits-cpu 5.0 ") == 0

        # Check that aioli model list shows only the latest version of all models
        # by default
        not_expected = (
            "my-model | the model description |         1 |       |"
            " fictional.registry.example/imagename"
        )
        expected = (
            "my-model | the model description |         2 |       |"
            " fictional.registry.example/imagename"
        )
        actual = subprocess.check_output(["aioli", "model", "list"]).decode("utf-8")
        assert actual.find(not_expected) < 0 and actual.find(expected) > 0

        # Check that aioli model list with the --all flag shows all versions of
        # all models
        expected = (
            "my-model | the model description |         2 |       |"
            " fictional.registry.example/imagename |\n my-model |"
            " the model description |         1 |       |"
            " fictional.registry.example/imagename"
        )

        actual = subprocess.check_output(["aioli", "models", "list", "--all"]).decode("utf-8")
        assert actual.find(expected) > 0

        # Create a different model
        assert (
            os.system(
                "aioli model create other-model "
                "--image fictional.registry.example/imagename "
                '--description "the model description" '
                "--requests-cpu 1.0 "
                "--requests-memory 1Gi "
                "--requests-gpu 1.0 "
                "--limits-cpu 2.5 "
                "--limits-memory 1Gi "
                "--limits-gpu 1.0 "
            )
            == 0
        )

        not_expected = (
            "other-model | the model description |         1 |       |"
            " fictional.registry.example/imagename"
        )
        actual = subprocess.check_output(["aioli", "models", "list", "--name", "my-model"]).decode(
            "utf-8"
        )
        # Same expected as previous
        assert actual.find(not_expected) < 0 and actual.find(expected) > 0

    # This test uses and cleans up models created in test_model_list
    def test_model_show(self, setup_login: None) -> None:
        # Check that aioli model show displays the latest version by default
        # when multiple versions of the model exist
        actual = subprocess.check_output(["aioli", "model", "show", "my-model"]).decode("utf-8")
        assert actual.find("version: 2") > 0 and actual.find("version: 1") < 0

        # Cleanup
        assert (
            os.system(
                "aioli model delete my-model.v1 && "
                "aioli model delete my-model.v2 && "
                "aioli model delete other-model"
            )
            == 0
        )

    def test_model_create(self, setup_login: None) -> None:
        # Create an entry for model table and test for expected values
        assert (
            os.system(
                "aioli model create iris-tf-keras --registry bento-registry1 "
                "--url s3://demo-bento-registry/iris-tf-keras "
                "--image fictional.registry.example/imagename "
                '--description "the model description" '
                "-e name1 --env name2=value2 "
                "-a arg1 --arg arg2 '-a -optarg' "
                "--requests-cpu 1.0 "
                "--requests-memory 1Gi "
                "--requests-gpu 1.0 "
                "--limits-cpu 2.5 "
                "--limits-memory 1Gi "
                "--limits-gpu 1.0 "
                "--gpu-type T4 "
                "--format openllm"
            )
            == 0
        )
        expected = (
            "iris-tf-keras | the model description |         1 | "
            "s3://demo-bento-registry/iris-tf-keras | "
            "fictional.registry.example/imagename | bento-registry1"
        )
        actual = subprocess.check_output(["aioli", "model", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

    def test_model_list_csv(self, setup_login: None) -> None:
        csv_output = subprocess.check_output(["aioli", "model", "list", "--csv"]).decode("utf-8")
        temp_file: str = "/tmp/test-cli-m.csv"
        with open(temp_file, "w") as file:
            file.write(csv_output)

        matching_row_found = False
        with open(temp_file, newline="") as file:
            reader = csv.DictReader(file)
            for row in reader:
                if (
                    row["Name"] == "iris-tf-keras"
                    and row["Description"] == "the model description"
                    and row["Version"] == "1"
                    and row["URI"] == "s3://demo-bento-registry/iris-tf-keras"
                    and row["Image"] == "fictional.registry.example/imagename"
                    and row["Registry"] == "bento-registry1"
                ):
                    matching_row_found = True
                    break

        assert matching_row_found, "No row with the expected properties was found"

    # Exercise various means of specifying the model version, using the show command as a vehicle.
    def test_model_versions(self, setup_login: None) -> None:
        yaml_output = subprocess.check_output(["aioli", "model", "show", "iris-tf-keras"]).decode(
            "utf-8"
        )
        temp_file: str = "/tmp/test-cli-m.yaml"

        # Demonstrate that we have valid yaml
        with open(temp_file, "w") as file:
            file.write(yaml_output)
        with open(temp_file, newline="") as file:
            from ruamel.yaml import YAML

            yaml = YAML(typ="safe")
            model = yaml.load(file)

        assert model["version"] == 1, "Validity check -- version == 1"

        yaml_output2 = subprocess.check_output(
            ["aioli", "model", "show", "iris-tf-keras.v1"]
        ).decode("utf-8")
        assert yaml_output == yaml_output2, "Expect the same output using .v1"

        yaml_output2 = subprocess.check_output(
            ["aioli", "model", "show", "iris-tf-keras.V1"]
        ).decode("utf-8")
        assert yaml_output == yaml_output2, "Expect the same output using .V1"

    # fmt: off
    def test_model_list_json_yaml(self, setup_login: None) -> None:
        expected = (
            '[\n'    # noqa: Q000
            '  {\n'  # noqa: Q000
            '    "arguments": [\n'
            '      "arg1",\n'
            '      "arg2",\n'
            '      "-optarg"\n'
            '    ],\n'  # noqa: Q000
            '    "cachingEnabled": false,\n'
            '    "description": "the model description",\n'
            '    "environment": {\n'
            '      "name1": "",\n'
            '      "name2": "value2"\n'
            '    },\n'  # noqa: Q000
            '    "image": "fictional.registry.example/imagename",\n'
            '    "modelFormat": "openllm",\n'
            '    "name": "iris-tf-keras",\n'
            '    "registry": "bento-registry1",\n'
            '    "resources": {\n'
            '      "gpuType": "T4",\n'
            '      "limits": {\n'
            '        "cpu": "2.5",\n'
            '        "gpu": "1.0",\n'
            '        "memory": "1Gi"\n'
            '      },\n'  # noqa: Q000
            '      "requests": {\n'
            '        "cpu": "1.0",\n'
            '        "gpu": "1.0",\n'
            '        "memory": "1Gi"\n'
            '      }\n'  # noqa: Q000
            '    },\n'   # noqa: Q000
            '    "url": "s3://demo-bento-registry/iris-tf-keras",\n'
            '    "version": 1\n'
            '  }\n'  # noqa: Q000
            ']\n'    # noqa: Q000
        )
        actual = subprocess.check_output(["aioli", "model", "list", "--json"]).decode("utf-8")
        assert actual.find(expected) >= 0

        expected_obj = json.loads(expected)
        actual_obj = yaml.safe_load(
            subprocess.check_output(["aioli", "model", "list", "--yaml"]).decode("utf-8"))

        assert actual_obj == expected_obj
    # fmt: on

    def test_model_create_no_image(self, setup_login: None) -> None:
        # The image is optional; verify...
        assert (
            os.system(
                "aioli model create openllm --registry bento-registry1 "
                "--format openllm "
                "--url s3://demo-bento-registry/iris-tf-keras "
                '--description "the model description"'
            )
            == 0
        )
        expected = (
            "openllm       | the model description |         1 | "
            "s3://demo-bento-registry/iris-tf-keras |"
            "                                      | bento-registry1"
        )
        actual = subprocess.check_output(["aioli", "model", "list"]).decode("utf-8")

        assert (actual.find(expected)) > 0
        assert os.system("aioli model delete openllm.v1") == 0

    def test_model_update_bad_model_name(self, setup_login: None) -> None:
        # Attempt to update model specifying the version incorrectly. Here we specify
        # a suffix of ".1", and not ".v1" which would specify the name & version.
        # assert (os.system("aioli model update iris-tf-keras.1") == 0)
        try:
            subprocess.check_output(
                ["aioli", "model", "update", "iris-tf-keras.1"], stderr=subprocess.STDOUT
            )
        except subprocess.CalledProcessError as e:
            assert e.returncode == 1
            expected = "Failed to modify a packaged model: model 'iris-tf-keras.1' does not exist."
            actual: str = e.output.decode("utf-8")
            assert actual.find(expected) == 0

    def test_model_update(self, setup_login: None) -> None:
        # Update existing model entry and test for expected values
        assert (
            os.system(
                "aioli model update iris-tf-keras --name iris-tf-keras1 "
                "--registry bento-registry1 "
                "--url s3://demo-bento-registry/iris-tf-keras_updated "
                "--image fictional.registry.example/updated_imagename "
                '--description "the updated model description"'
            )
            == 0
        )
        actual = subprocess.check_output(["aioli", "model", "list", "--all"]).decode("utf-8")
        expected = (
            "iris-tf-keras1 | the updated model description |         1 "
            "| s3://demo-bento-registry/iris-tf-keras_updated | "
            "fictional.registry.example/updated_imagename | bento-registry1"
        )
        assert (actual.find(expected)) > 0

        # Create a second version of the model
        assert (
            os.system(
                "aioli model update iris-tf-keras --name iris-tf-keras "
                "--registry bento-registry1 "
                "--url s3://demo-bento-registry/iris-tf-keras_updated_v2 "
                "--image fictional.registry.example/updated_imagename "
                '--description "the updated model description"'
            )
            == 0
        )
        actual = subprocess.check_output(["aioli", "model", "list", "--all"]).decode("utf-8")
        expected = (
            "iris-tf-keras  | the updated model description |         2 "
            "| s3://demo-bento-registry/iris-tf-keras_updated_v2 | "
            "fictional.registry.example/updated_imagename | bento-registry1"
        )
        assert (actual.find(expected)) > 0

        # Try to create a third version of the model which fails without version
        assert (
            os.system(
                "aioli model update iris-tf-keras --name iris-tf-keras "
                "--registry bento-registry1 "
                "--url s3://demo-bento-registry/iris-tf-keras_updated_v2 "
                "--image fictional.registry.example/updated_imagename "
                '--description "the updated model description"'
            )
            == 256
        )

        # Create a third version of the model using the second version
        assert (
            os.system(
                "aioli model update iris-tf-keras.v2 --name iris-tf-keras "
                "--registry bento-registry1 "
                "--url s3://demo-bento-registry/iris-tf-keras_updated_v3 "
                "--image fictional.registry.example/updated_imagename "
                '--description "the updated model description"'
            )
            == 0
        )

        actual = subprocess.check_output(["aioli", "model", "list", "--all"]).decode("utf-8")

        expected = (
            "iris-tf-keras  | the updated model description |         3 "
            "| s3://demo-bento-registry/iris-tf-keras_updated_v3 | "
            "fictional.registry.example/updated_imagename | bento-registry1"
        )
        assert (actual.find(expected)) > 0

        # Create a fourth with a new name of the model using the second version
        # (specifying the version with the optional v{n} format)
        assert (
            os.system(
                "aioli model update iris-tf-keras.v2 --name iris-tf-keras-v4 "
                "--registry bento-registry1 "
                "--url s3://demo-bento-registry/iris-tf-keras_updated_v3 "
                "--image fictional.registry.example/updated_imagename "
                '--description "the updated model description"'
            )
            == 0
        )

        actual = subprocess.check_output(["aioli", "model", "list", "--all"]).decode("utf-8")

        expected = (
            "iris-tf-keras-v4 | the updated model description |         1 "
            "| s3://demo-bento-registry/iris-tf-keras_updated_v3 | "
            "fictional.registry.example/updated_imagename | bento-registry1"
        )
        assert (actual.find(expected)) > 0

    def test_deployment_create(self, setup_login: None) -> None:
        # Create a deployment and test for expected values
        assert (
            os.system(
                "aioli deployment create --model iris-tf-keras "
                "--namespace aioli "
                "--authentication-required false "
                "iris-tf-keras-deployment"
            )
            == 0
        )
        actual = subprocess.check_output(["aioli", "deployment", "list"]).decode("utf-8")
        expected = (
            "iris-tf-keras-deployment | iris-tf-keras |         3 | aioli       "
            "| Deploying | False           | Deploying |           0"
        )

        assert actual.find(expected) > 0

    def test_deployment_list_csv(self, setup_login: None) -> None:
        csv_output = subprocess.check_output(["aioli", "deployment", "list", "--csv"]).decode(
            "utf-8"
        )
        temp_file: str = "/tmp/test-cli-d.csv"
        with open(temp_file, "w") as file:
            file.write(csv_output)
        with open(temp_file, newline="") as file:
            reader = csv.DictReader(file)
            for row in reader:
                assert (
                    row["Name"] == "iris-tf-keras-deployment"
                ), "Expected iris-tf-keras-deployment"
                assert row["Model"] == "iris-tf-keras", "Expected iris-tf-keras"
                assert row["Namespace"] == "aioli", "Expected aioli"
                assert row["Status"] == "Deploying", "Expected Deploying"
                assert row["Auth Required"] == "False", "Expected False"
                assert row["State"] == "Deploying", "Expected Deploying"
                assert row["Traffic %"] == "0", "Expected 0"

    def test_deployment_update(self, setup_login: None) -> None:
        # Update the deployment and test for expected values.
        assert (
            os.system(
                "aioli deployment update "
                "--authentication-required true "
                "--model iris-tf-keras "
                "iris-tf-keras-deployment"
            )
            == 0
        )
        actual = subprocess.check_output(["aioli", "deployment", "list"]).decode("utf-8")

        expected = (
            "iris-tf-keras-deployment | iris-tf-keras |         3 | aioli       "
            "| Deploying | True            | Deploying |           0"
        )
        assert actual.find(expected) > 0

        # Disallow --pause and --resume at the same time
        proc = subprocess.run(
            [
                "aioli",
                "deployment",
                "update",
                "iris-tf-keras-deployment",
                "--pause",
                "--resume",
            ],
            capture_output=True,
        )
        assert proc.returncode == 1
        expected = "--pause and --resume cannot be specified at the same time"
        assert proc.stderr.decode("utf-8").find(expected) == 0

        # List the models and it's version for the deployment
        actual = subprocess.check_output(
            ["aioli", "model", "list-deployments", "iris-tf-keras"]
        ).decode("utf-8")
        expected = "| iris-tf-keras |         3 |       100"
        assert actual.find(expected) > 0

    # fmt: off
    def test_deployment_list_json_yaml(self, setup_login: None) -> None:
        expected = (
            '[\n'    # noqa: Q000
            '  {\n'  # noqa: Q000
            '    "arguments": [],\n'
            '    "autoScaling": {\n'
            '      "maxReplicas": 1,\n'
            '      "metric": "concurrency",\n'
            '      "minReplicas": 0,\n'
            '      "target": 1\n'
            '    },\n'  # noqa: Q000
            '    "canaryTrafficPercent": 100,\n'
            '    "environment": {},\n'
            '    "goalStatus": "Ready",\n'
            '    "model": "iris-tf-keras",\n'
            '    "name": "iris-tf-keras-deployment",\n'
            '    "namespace": "aioli",\n'
            '    "nodeSelectors": {},\n'
            '    "priorityClassName": "",\n'
            '    "secondaryState": {\n'
            '      "endpoint": "",\n'
            '      "modelId": "",\n'
            '      "nativeAppName": "",\n'
            '      "status": "None",\n'
            '      "trafficPercentage": 0\n'
            '    },\n'  # noqa: Q000
            '    "security": {\n'
            '      "authenticationRequired": true\n'
            '    },\n'  # noqa: Q000
            '    "state": {\n'
            '      "endpoint": "",\n'
            '      "modelId": "",\n'
            '      "nativeAppName": "",\n'
            '      "status": "Deploying",\n'
            '      "trafficPercentage": 0\n'
            '    },\n'  # noqa: Q000
            '    "status": "Deploying",\n'
            '    "version": "3"\n'
            '  }\n'  # noqa: Q000
            ']\n'    # noqa: Q000
        )
        actual = subprocess.check_output(
            ["aioli", "deployment", "list", "--json"]).decode("utf-8")
        assert actual == expected

        expected_obj = json.loads(expected)
        actual_obj = yaml.safe_load(subprocess.check_output(
            ["aioli", "deployment", "list", "--yaml"]).decode("utf-8"))
        assert actual_obj == expected_obj
    # fmt: on

    def test_deployment_create_with_all_options(self, setup_login: None) -> None:
        # Create a second deployment with all supported options
        # and test for expected values
        assert (
            os.system(
                "aioli deployment create --model iris-tf-keras "
                "--namespace aioli "
                "--authentication-required false "
                "--autoscaling-min-replicas 1 "
                "--autoscaling-max-replicas 10 "
                "--autoscaling-target 1 "
                "--autoscaling-metric concurrency "
                "--canary-traffic-percent 20 "
                "-a='--debug' "
                "-e MODS=SOME "
                "--node-selector kubernetes.io/arch=amd64 "
                "--priority-class-name '' "
                "iris-tf-keras-deployment-2"
            )
            == 0
        )

    def test_deployment_show(self, setup_login: None) -> None:
        # Test deployment table header, there are no entries at this point.
        assert os.system("aioli deployment show iris-tf-keras-deployment-2") == 0
        expected = None
        data_file = self.testdir + "/aioli_deployment_show.txt"
        with open(data_file, "r") as file:
            expected = file.read()
        result = subprocess.check_output(
            ["aioli", "deployment", "show", "iris-tf-keras-deployment-2"]
        ).decode("utf-8")
        result_list = result.split("\n")
        filtered_result = ""

        # Filter out modifiedAt and id fields as they vary for each test run
        for line in result_list:
            if "modifiedAt" in line:
                continue
            if "id" in line:
                continue
            filtered_result = filtered_result + "\n" + line

        actual = filtered_result.strip()
        assert actual == expected

    def test_deployment_update_args_environment(self, setup_login: None) -> None:
        # Update the deployments arguments and environment and check that changes
        # were made.
        assert (
            os.system(
                "aioli deployment update "
                "-a='--updated' "
                "-e MODS=UPDATED "
                "--env OTHER=VALUE "
                "iris-tf-keras-deployment-2"
            )
            == 0
        )

        result = subprocess.check_output(
            ["aioli", "deployment", "show", "iris-tf-keras-deployment-2"]
        ).decode("utf-8")
        result_list = result.split("\n")
        args_found = False
        other_found = False
        node_selector_found = False
        for line in result_list:
            if "- --updated" in line:
                args_found = True
            if "MODS: UPDATED" in line:
                mods_found = True
            if "OTHER: VALUE" in line:
                other_found = True
            if "kubernetes.io/arch: amd64" in line:
                node_selector_found = True

        assert args_found
        assert mods_found
        assert other_found
        assert node_selector_found

    def test_deployment_update_node_selectors(self, setup_login: None) -> None:
        # Update the deployments nodeSelectors check that changes
        # were made. Also verify arguments and environment are preserved.
        assert (
            os.system(
                "aioli deployment update "
                "--node-selector x/y=a-b "
                "--node-selector x/z=a-c "
                "iris-tf-keras-deployment-2"
            )
            == 0
        )

        result = subprocess.check_output(
            ["aioli", "deployment", "show", "iris-tf-keras-deployment-2"]
        ).decode("utf-8")
        result_list = result.split("\n")
        args_found = False
        other_found = False
        prev_node_selector_not_found = True
        node_selector_1_found = False
        node_selector_2_found = False
        for line in result_list:
            if "- --updated" in line:
                args_found = True
            if "MODS: UPDATED" in line:
                mods_found = True
            if "OTHER: VALUE" in line:
                other_found = True
            if "kubernetes.io/arch: amd64" in line:
                prev_node_selector_not_found = False
            if "x/y: a-b" in line:
                node_selector_1_found = True
            if "x/z: a-c" in line:
                node_selector_2_found = True
        assert args_found
        assert mods_found
        assert other_found
        assert node_selector_1_found
        assert node_selector_2_found
        assert prev_node_selector_not_found

    def test_deployment_clear_env_args_node_selectors(self, setup_login: None) -> None:
        # Update the deployments and clear env/args/node_selectors
        assert (
            os.system(
                "aioli deployment update "
                "iris-tf-keras-deployment-2 "
                "--env "
                "--arg "
                "--node-selector"
            )
            == 0
        )

        result = subprocess.check_output(
            ["aioli", "deployment", "show", "iris-tf-keras-deployment-2"]
        ).decode("utf-8")
        result_list = result.split("\n")
        args_not_found = True
        mods_not_found = True
        other_not_found = True
        node_selector_1_not_found = True
        node_selector_2_not_found = True
        for line in result_list:
            if "- --updated" in line:
                args_not_found = False
            if "MODS: UPDATED" in line:
                mods_not_found = False
            if "OTHER: VALUE" in line:
                other_not_found = False
            if "x/y: a-b" in line:
                node_selector_1_not_found = False
            if "x/z: a-c" in line:
                node_selector_2_not_found = False
        assert args_not_found
        assert mods_not_found
        assert other_not_found
        assert node_selector_1_not_found
        assert node_selector_2_not_found

    def test_deployment_delete(self, setup_login: None) -> None:
        assert os.system("aioli deployment delete iris-tf-keras-deployment") == 0
        assert os.system("aioli deployment delete iris-tf-keras-deployment-2") == 0

    def test_model_delete(self, setup_login: None) -> None:
        assert os.system("aioli model delete iris-tf-keras1") == 0

        # Test that a model with multiple versions cannot be deleted by only name
        assert os.system("aioli model delete iris-tf-keras") != 0

        assert os.system("aioli model delete iris-tf-keras.v1") == 0
        assert os.system("aioli model delete iris-tf-keras.v2") == 0
        assert os.system("aioli model delete iris-tf-keras.v3") == 0
        assert os.system("aioli model delete iris-tf-keras-v4.v1") == 0

    def test_registry_delete(self, setup_login: None) -> None:
        # Delete the registry entry
        assert os.system("aioli registry delete bento-registry1") == 0

    def test_user_list(self, setup_login: None) -> None:
        # Test the user table header fields
        assert os.system("aioli user list") == 0
        expected = "admin      | Default Administrator | True     | False"
        actual = subprocess.check_output(["aioli", "user", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

    def test_user_list_json_yaml(self, setup_login: None) -> None:
        expected = (
            "[\n"
            "  {\n"
            '    "active": true,\n'
            '    "displayName": "Default Administrator",\n'
            '    "id": "1",\n'
            '    "remote": false,\n'
            '    "username": "admin"\n'
            "  }\n"
            "]\n"
        )

        expected = json.loads(expected)

        actual = json.loads(
            subprocess.check_output(["aioli", "user", "list", "--json"]).decode("utf-8")
        )
        for d in actual:
            del d["lastAuthAt"]
            del d["modifiedAt"]
        assert actual == expected

        actual = yaml.safe_load(
            subprocess.check_output(["aioli", "user", "list", "--yaml"]).decode("utf-8")
        )
        for d in actual:
            del d["lastAuthAt"]
            del d["modifiedAt"]
        assert actual == expected

    def test_user_list_csv(self, setup_login: None) -> None:
        csv_output = subprocess.check_output(["aioli", "user", "list", "--csv"]).decode("utf-8")
        temp_file: str = "/tmp/test-cli-u.csv"
        with open(temp_file, "w") as file:
            file.write(csv_output)
        with open(temp_file, newline="") as file:
            reader = csv.DictReader(file)
            for row in reader:
                assert row["Username"] == "admin", "Expected admin"
                assert (
                    row["Display Name"] == "Default Administrator"
                ), "Expected Default Administrator"
                assert row["Active"] == "True", "Expected True"
                assert row["Remote"] == "False", "Expected False"

    def test_user_create(self, setup_login: None) -> None:
        # create a user and test for expected values
        assert os.system("aioli user create testuser --password passwd") == 0
        expected = "testuser   |"
        actual = subprocess.check_output(["aioli", "user", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

    def test_user_update(self, setup_login: None) -> None:
        # update testuser and check for expected values
        assert (
            os.system(
                "aioli user update testuser --username newuser --display-name 'new display name'"
            )
            == 0
        )
        expected = "newuser    | new display name      | True"
        actual = subprocess.check_output(["aioli", "user", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0
        assert os.system("aioli user update newuser --username testuser --active f") == 0
        expected = "testuser   | new display name      | False"
        actual = subprocess.check_output(["aioli", "user", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

    def test_user_whoami(self, setup_login: None) -> None:
        assert os.system("aioli user whoami") == 0
        expected = "You are logged in as user 'admin'\n"
        actual = subprocess.check_output(["aioli", "user", "whoami"]).decode("utf-8")
        assert actual == expected

    def test_user_activate(self, setup_login: None) -> None:
        # Check activate feature
        assert os.system("aioli user activate testuser") == 0

    def test_user_deactivate(self, setup_login: None) -> None:
        # Check deactivate feature
        assert os.system("aioli user deactivate testuser") == 0

    def test_user_create_remote(self, setup_login: None) -> None:
        # create a remote user and test for expected values
        assert os.system("aioli user create --remote testuser-r") == 0
        expected = "testuser-r |                       | True     | True"
        actual = subprocess.check_output(["aioli", "user", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

    def test_template_resource_create_and_update(self, setup_login: None) -> None:
        # Create a resource template
        assert (
            os.system(
                "aioli template resource create my-template "
                '--description "My test reource template" '
                "--gpu-type nvidia "
                "--limits-cpu 10 "
                "--limits-memory 10Gi "
                "--limits-gpu 4 "
                "--requests-cpu 0 "
                "--requests-memory 10Gi "
                "--requests-gpu 4 "
            )
            == 0
        )

        expected = (
            "my-template | My test reource template               | Request: 0, Limit: 10 |"
            " Request: 10Gi, Limit: 10Gi | Request: 4, Limit: 4 | nvidia"
        )
        actual = subprocess.check_output(["aioli", "template", "resource", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

        # Make an update where gpu_type remains unchanged
        assert (
            os.system(
                "aioli template resource update my-template "
                '--description "My updated test resource template" '
                "--limits-cpu 5 "
                "--limits-memory 100Gi "
                "--limits-gpu 0 "
                "--requests-cpu 5 "
                "--requests-memory 100Gi "
                "--requests-gpu 0 "
            )
            == 0
        )

        expected = (
            "my-template | My updated test resource template      | Request: 5, Limit: 5  |"
            " Request: 100Gi, Limit: 100Gi | Request: 0, Limit: 0 | nvidia"
        )
        actual = subprocess.check_output(["aioli", "template", "resource", "list"]).decode("utf-8")
        assert (actual.find(expected)) > 0

    def test_template_autoscaling_create_and_update(self, setup_login: None) -> None:
        # Create a resource template
        assert (
            os.system(
                "aioli template autoscaling create my-template "
                '--description "My test autoscaling template" '
                "--autoscaling-min-replicas 0 "
                "--autoscaling-max-replicas 2 "
                "--autoscaling-metric rps "
                "--autoscaling-target 25 "
            )
            == 0
        )

        expected = (
            "my-template                | My test autoscaling template            "
            "                       |"
            "             0 |             2 | rps         |       25"
        )
        actual = subprocess.check_output(["aioli", "template", "autoscaling", "list"]).decode(
            "utf-8"
        )
        assert (actual.find(expected)) > 0

        # Make an update where autoscaling_metric remains unchanged
        assert (
            os.system(
                "aioli template autoscaling update my-template "
                '--description "My updated test autoscaling template" '
                "--autoscaling-min-replicas 2 "
                "--autoscaling-max-replicas 2 "
                "--autoscaling-target 50 "
            )
            == 0
        )

        expected = (
            "my-template                | My updated test autoscaling template              "
            "             |             2 |             2 | rps         |       50"
        )

        actual = subprocess.check_output(["aioli", "template", "autoscaling", "list"]).decode(
            "utf-8"
        )
        assert (actual.find(expected)) > 0

    def test_template_resource_show(self, setup_login: None) -> None:
        assert os.system("aioli template resource show my-template") == 0

        expected = yaml.safe_load(
            "description: My updated test resource template\n"
            "name: my-template\n"
            "resources:\n"
            "  gpuType: nvidia\n"
            "  limits:\n"
            "    cpu: '5'\n"
            "    gpu: '0'\n"
            "    memory: 100Gi\n"
            "  requests:\n"
            "    cpu: '5'\n"
            "    gpu: '0'\n"
            "    memory: 100Gi\n"
        )
        actual = yaml.safe_load(
            subprocess.check_output(["aioli", "template", "resource", "show", "my-template"])
        )
        del actual["modifiedAt"]
        assert actual == expected

    def test_template_autoscaling_show(self, setup_login: None) -> None:
        assert os.system("aioli template autoscaling show my-template") == 0

        expected = yaml.safe_load(
            "autoScaling:\n"
            "  maxReplicas: 2\n"
            "  metric: rps\n"
            "  minReplicas: 2\n"
            "  target: 50\n"
            "description: My updated test autoscaling template\n"
            "name: my-template\n"
        )
        actual = yaml.safe_load(
            subprocess.check_output(["aioli", "template", "autoscaling", "show", "my-template"])
        )
        del actual["modifiedAt"]
        assert actual == expected

    def test_template_resource_delete(self, setup_login: None) -> None:
        assert os.system("aioli template resource delete my-template") == 0

    def test_template_autoscaling_delete(self, setup_login: None) -> None:
        assert os.system("aioli template autoscaling delete my-template") == 0



# coding: utf-8

"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictBool, StrictInt, StrictStr
from pydantic import Field
from aiolirest.models.configuration_resources import ConfigurationResources
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class PackagedModel(BaseModel):
    """
    The PackagedModel describes a specific model that can be loaded. It optionally references a model registry to provide authorization, and other details.
    """ # noqa: E501
    arguments: Optional[List[StrictStr]] = Field(default=None, description="Arguments to be added to the service command line")
    caching_enabled: Optional[StrictBool] = Field(default=None, description="Provide cache-on-first-use of the model download. Only users with the Admin role have permission to activate this setting.", alias="cachingEnabled")
    description: Optional[StrictStr] = Field(default=None, description="A description of the model.")
    environment: Optional[Dict[str, StrictStr]] = Field(default=None, description="Environment variables added to the service")
    id: Optional[StrictStr] = Field(default=None, description="The ID of the model.  This is a read-only field and is automatically assigned on creation.")
    image: Optional[StrictStr] = Field(default=None, description="Alternate Docker container image servicing the model. Required for a nim or custom model format. Leave blank to use the default image.")
    format: Optional[StrictStr] = Field(default='custom', description="Model format for downloaded models (e.g. from s3, http, etc.). Must be one of the values: (bento-archive, openllm, nim, custom). * `custom` - The packaged model is provided in a container image. * `bento-archive` - The packaged model is a bento archive file (.bento).  It will be downloaded,  expanded, and then  will be served using the `bentoml serve` command in a provided bentoml serving container. * `openllm` - The packaged model will be served using the `openllm serve` command in a  provided openllm serving container. * `vllm` - The packaged model will be served using the vllm/openai-vllm container image. * `nim` - The packaged model will be served using the specified NVIDIA NIM microservices container image.", alias="modelFormat")
    modified_at: Optional[StrictStr] = Field(default=None, description="Date-time of last modification of the model. This is a read-only field and is automatically updated.", alias="modifiedAt")
    name: StrictStr = Field(description="The name of the model.  Must begin with a letter, but may contain letters, numbers, and hyphen.")
    project: Optional[StrictStr] = Field(default=None, description="The ID of the containing project")
    registry: Optional[StrictStr] = Field(default=None, description="The name or ID of the model registry.")
    resources: Optional[ConfigurationResources] = None
    url: Optional[StrictStr] = Field(default=None, description="Model data required by the deployment.  Supported schemes are:  * `openllm://` - An openllm compatible model dynamically downloaded from huggingface.co. This is only supported for the openllm model format. * `hf://` - A vLLM compatible model dynamically downloaded from huggingface.co. This is only supported for the vllm model format. * `s3://` - A model directory dynamically downloaded from an associated s3  registry bucket.  This is supported for the bento-archive, openllm, vllm, and custom model formats. * `pvc://` - A persistent volume claim (PVC) mounted by the deployment where models or data can be stored.  This is supported for the NIM or custom model formats. The syntax is `pvc://<pvc-name>/<path>` with URI parameters `?containerPath=<container-path>` to specify where in the container the PVC should be mounted, and ` ?readOnly` to perform the mount in read-only mode.  The `<pvc-name>` must already exist in the kubernetes namespace where the packaged model will be deployed.   If the optional `<path>` is specified, the specified path within the PVC will be mounted. * `pfs://` - A model directory directly downloaded from PFS.  This is supported for the bento-archive, openllm, vllm, and custom model formats.  The syntax is `pfs://<project>/<repo>@<commit>[:<path>]` with the optional URI parameter `?containerPath=<container-path>` to specify where in the container the model directory should be mounted.")
    version: Optional[StrictInt] = Field(default=None, description="The version of the model.  This is a read-only field and is automatically assigned on creation.")
    __properties: ClassVar[List[str]] = ["arguments", "cachingEnabled", "description", "environment", "id", "image", "modelFormat", "modifiedAt", "name", "project", "registry", "resources", "url", "version"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of PackagedModel from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of resources
        if self.resources:
            _dict['resources'] = self.resources.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of PackagedModel from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "arguments": obj.get("arguments"),
            "cachingEnabled": obj.get("cachingEnabled"),
            "description": obj.get("description"),
            "environment": obj.get("environment"),
            "id": obj.get("id"),
            "image": obj.get("image"),
            "modelFormat": obj.get("modelFormat") if obj.get("modelFormat") is not None else 'custom',
            "modifiedAt": obj.get("modifiedAt"),
            "name": obj.get("name"),
            "project": obj.get("project"),
            "registry": obj.get("registry"),
            "resources": ConfigurationResources.from_dict(obj.get("resources")) if obj.get("resources") is not None else None,
            "url": obj.get("url"),
            "version": obj.get("version")
        })
        return _obj





# © Copyright 2023-2024 Hewlett Packard Enterprise Development LP
import re
from argparse import Namespace
from typing import Any, Dict, List, Optional

from pydantic import StrictInt

import aiolirest
from aioli import cli
from aioli.cli import render
from aioli.cli.registry import lookup_registry_name_by_id
from aioli.common import api
from aioli.common.api import authentication
from aioli.common.api.errors import NotFoundException
from aioli.common.declarative_argparse import Arg, ArgsDescription, Cmd, Group
from aioli.common.util import (
    construct_arguments,
    construct_environment,
    launch_dashboard,
)
from aiolirest.models.configuration_resources import ConfigurationResources
from aiolirest.models.deployment_model_version import DeploymentModelVersion
from aiolirest.models.packaged_model import PackagedModel
from aiolirest.models.packaged_model_request import PackagedModelRequest
from aiolirest.models.resource_profile import ResourceProfile


@authentication.required
def list_models(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.PackagedModelsApi(session)
        response = None
        # Get all models
        if args.all:
            response = api_instance.models_get()
        # Get all versions of specified model
        elif args.name is not None:
            response = api_instance.models_get(name=args.name)
        # Get latest version of each model (default)
        else:
            response = api_instance.models_get(latest="")

    registries_api = aiolirest.RegistriesApi(session)

    if args.json:
        render.print_json(format_json(response, registries_api))
    elif args.yaml:
        print(render.format_object_as_yaml(format_json(response, registries_api)))
    else:
        format_models(response, args, registries_api)


def format_json(
    response: List[PackagedModel], registries_api: aiolirest.RegistriesApi
) -> List[Dict[str, str]]:
    models = []
    for m in response:
        # Don't use the m.to_json() method as it adds backslash escapes for double quote
        d = m.to_dict()
        d.pop("id")
        d.pop("modifiedAt")
        d.pop("project")  # something for INF-1516
        rname = lookup_registry_name_by_id(m.registry, registries_api)
        d["registry"] = rname
        models.append(d)

    return models


def format_models(
    response: List[PackagedModel], args: Namespace, registries_api: aiolirest.RegistriesApi
) -> None:
    def format_model(e: PackagedModel, reg_api: aiolirest.RegistriesApi) -> List[Any]:
        rname = lookup_registry_name_by_id(e.registry, reg_api)
        result = [
            e.name,
            e.description,
            e.version,
            e.url,
            e.image,
            rname,
        ]
        return result

    headers = [
        "Name",
        "Description",
        "Version",
        "URI",
        "Image",
        "Registry",
    ]
    values = [format_model(r, registries_api) for r in response]
    render.tabulate_or_csv(headers, values, args.csv)


@authentication.required
def create(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.PackagedModelsApi(session)
        requests = ResourceProfile(
            cpu=args.requests_cpu, gpu=args.requests_gpu, memory=args.requests_memory
        )
        limits = ResourceProfile(
            cpu=args.limits_cpu, gpu=args.limits_gpu, memory=args.limits_memory
        )
        resources = ConfigurationResources(gpuType=args.gpu_type, requests=requests, limits=limits)
        r = PackagedModelRequest(
            name=args.name,
            description=args.description,
            url=args.url,
            image=args.image,
            registry=args.registry,
            resources=resources,
            environment=construct_environment(args),
            modelFormat=args.format,
            arguments=construct_arguments(args),
        )

        if args.enable_caching:
            r.caching_enabled = True

        if args.disable_caching:
            r.caching_enabled = False

        api_instance.models_post(r)


def get_models(name: str, api: aiolirest.PackagedModelsApi) -> List[PackagedModel]:
    """
    Fetch a list of PackagedModel objects matching a given name.

    Args:
       name: The name of the model. The version can also be optionally specified
         with the model name. It can be represented as <model_name>.[vV]n
         example: a model whose name is fb125m-model and version is 1 can be
         represented as fb125m-model.v1 or fb125m-model.V1
         When the version is passed with the model name, it will be parsed from the
         name and lookup done for deployments to match the model name and version.
       api: An API object used to make calls to the controller.

    Returns:
       List of PackagedModel objects matching the model name. The version can be
       passed implicitly along with the model name. If it is passed implicitly,
       the version is parsed out from the name and used to lookup for matching
       deployments.

    Raises:
       NotFoundException: If no model is found that matches the name of the
         model, an exception will be raised. If the version is implicitly passed,
         and no models are found after parsing out the version from the model
         name, an exception will be raised.
    """
    packaged_models: List[PackagedModel] = api.models_get()

    # Check if the version suffix was provided in the model name
    version: Optional[str] = None
    m = re.match(r"^(.+)\.[Vv](\d+)$", name)
    if m:
        name = m.group(1)
        version = m.group(2)
    models = [r for r in packaged_models if r.name == name]

    # If list is empty, we did not find models matching the criteria.
    if len(models) == 0:
        raise NotFoundException(f"model {name} not found")

    # if there is an explicit version specified, then use that
    if version:
        for r in models:
            if r.name == name:
                if r.version == StrictInt(version):
                    return [r]
        raise NotFoundException(f"model {name} version {version} not found")

    # If no version suffix is provided, we return a list all the model versions
    # for the deployment.
    return models


def lookup_model(name: str, api: aiolirest.PackagedModelsApi) -> PackagedModel:
    # From the database, get the model record. If the model exists in multiple versions,
    # then sufficient version information must be part of the request.
    models: List[PackagedModel] = api.models_get(name)
    if len(models):
        return models[0]

    raise NotFoundException(
        f"model {name} not found. Model versions may optionally be specified "
        "using the suffix '.v#', for example, '.v1', '.v100'"
    )


@authentication.required
def dashboard(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.PackagedModelsApi(session)

    model = lookup_model(args.name, api_instance)

    assert model.id is not None
    observability = api_instance.models_id_observability_get(model.id)
    launch_dashboard(args, observability.dashboard_url)


@authentication.required
def show_model(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.PackagedModelsApi(session)

    # By default, show latest (maximum) version
    model = api_instance.models_get(model_ref=args.name, latest="")[0]
    registries_api = aiolirest.RegistriesApi(session)

    rname = lookup_registry_name_by_id(model.registry, registries_api)

    d = model.to_dict()
    d["registry"] = rname
    d.pop("project")  # something for INF-1516

    if args.json:
        render.print_json(d)
    else:
        print(render.format_object_as_yaml(d))


@authentication.required
def update(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.PackagedModelsApi(session)
        found = lookup_model(args.modelname, api_instance)
        request = PackagedModelRequest(
            description=found.description,
            image=found.image,
            name=found.name,
            registry=found.registry,
            url=found.url,
            arguments=found.arguments,
            resources=found.resources,
            environment=found.environment,
            modelFormat=found.format,
            cachingEnabled=found.caching_enabled,
        )

        if (
            request.resources is None
            or request.resources.requests is None
            or request.resources.limits is None
        ):
            # Not likely, but testing these prevents complaints from mypy
            raise api.errors.BadResponseException("Unexpected null result")

        if args.limits_gpu is not None and args.requests_gpu is None:
            args.requests_gpu = args.limits_gpu

        if args.limits_gpu is None and args.requests_gpu is not None:
            args.limits_gpu = args.requests_gpu

        if args.name is not None:
            request.name = args.name

        if args.description is not None:
            request.description = args.description

        if args.url is not None:
            request.url = args.url

        if args.image is not None:
            request.image = args.image

        if args.registry is not None:
            request.registry = args.registry

        if args.format is not None:
            request.format = args.format

        if args.requests_cpu is not None:
            request.resources.requests.cpu = args.requests_cpu

        if args.requests_memory is not None:
            request.resources.requests.memory = args.requests_memory

        if args.requests_gpu is not None:
            request.resources.requests.gpu = args.requests_gpu

        if args.limits_cpu is not None:
            request.resources.limits.cpu = args.limits_cpu

        if args.limits_memory is not None:
            request.resources.limits.memory = args.limits_memory

        if args.limits_gpu is not None:
            request.resources.limits.gpu = args.limits_gpu

        if args.gpu_type is not None:
            request.resources.gpu_type = args.gpu_type

        if args.env is not None:
            request.environment = construct_environment(args)

        if args.arg is not None:
            request.arguments = construct_arguments(args)

        if args.enable_caching:
            request.caching_enabled = True

        if args.disable_caching:
            request.caching_enabled = False

        headers = {"Content-Type": "application/json"}

        assert found.id is not None
        api_instance.models_id_put(found.id, request, _headers=headers)


@authentication.required
def delete_model(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.PackagedModelsApi(session)
        found = lookup_model(args.name, api_instance)

        assert found.id is not None
        api_instance.models_id_delete(found.id)


@authentication.required
def auth_token(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.PackagedModelsApi(session)
        found = lookup_model(args.name, api_instance)
        assert found.id is not None
        response = api_instance.models_id_token_get(found.id)
    t = response.to_dict()
    print(render.format_object_as_yaml(t))


@authentication.required
def list_deployments(args: Namespace) -> None:
    with cli.setup_session(args) as session:
        api_instance = aiolirest.PackagedModelsApi(session)

        # For listing models, if no model version with model name,
        # we want to list all the versions of the deployed model
        models = get_models(args.name, api_instance)

        response: List[DeploymentModelVersion] = []
        for m in models:
            for d in api_instance.models_versions_get(m.id):
                response.append(d)

    def format_deployments(e: DeploymentModelVersion) -> List[Any]:
        result = [
            e.deployed,
            e.native_app_name,
            e.model,
            e.mdl_version,
            e.canary_traffic_percent,
        ]
        return result

    headers = [
        "Deployed",
        "Native App Name",
        "Model",
        "Model\nVersion",
        "Traffic",
    ]

    if args.json:
        render.print_json([r.to_dict() for r in response])
    elif args.yaml:
        print(render.format_object_as_yaml([r.to_dict() for r in response]))
    else:
        values = [format_deployments(r) for r in response]
        render.tabulate_or_csv(headers, values, args.csv)


common_model_args: ArgsDescription = [
    Arg("--description", help="Description of the packaged model"),
    Arg("--url", help="Reference within the specified registry"),
    Arg("--registry", help="The name or ID of the packaged model registry"),
    Arg(
        "--format",
        "--modelformat",
        help="Model format for downloaded models (bento-archive, openllm, nim, unspecified)",
    ),
    Group(
        Arg("--enable-caching", action="store_true", help="Enable caching for the packaged model"),
        Arg(
            "--disable-caching", action="store_true", help="Disable caching for the packaged model"
        ),
    ),
    Arg(
        "-a",
        "--arg",
        nargs="?",
        help="Argument to be added to the service command line. "
        "If specifying an argument that starts with a '-', use the form --arg=<your-argument>. "
        "Specifying any --arg replaces prior args with the arguments on this invocation. "
        "Use a single --arg with no value to clear all arguments.",
        action="append",
    ),
    Arg(
        "-e",
        "--env",
        nargs="?",
        help="Specifies an environment variable & value as name=value, "
        "to be passed to the launched container. "
        "Specifying any --env replaces prior environment vars with those on this invocation. "
        "Use a single --env with no value to clear all environment vars.",
        action="append",
    ),
    Arg("--gpu-type", help="GPU type required"),
    Arg("--limits-cpu", help="CPU limit"),
    Arg("--limits-memory", help="Memory limit"),
    Arg("--limits-gpu", help="GPU limit"),
    Arg("--requests-cpu", help="CPU request"),
    Arg("--requests-memory", help="Memory request"),
    Arg("--requests-gpu", help="GPU request"),
]

VERSIONED_MODEL_HELP_MSG = (
    "The packaged model id, name or versioned-name (evaluated in that order). "
    "A versioned-name is the package model name with suffix of the version "
    "with the format 'name.V###' where '###' is the version number. For example, "
    "a model named 'my-model' with a version of '23' would be represented by "
    "versioned-name of: my-model.V23"
)

main_cmd = Cmd(
    "m|odel|s",
    None,
    "manage packaged models",
    [
        # Inspection commands.
        Cmd(
            "list ls",
            list_models,
            "list only the latest version of each packaged model",
            [
                Group(
                    Arg("--all", action="store_true", help="show all models"),
                    Arg("--name", help="list all versions of model, if specified"),
                ),
                Group(
                    Arg("--csv", action="store_true", help="print as CSV"),
                    Arg("--json", action="store_true", help="print as JSON"),
                    Arg("--yaml", action="store_true", help="print as YAML"),
                ),
            ],
            is_default=True,
        ),
        # Create command.
        Cmd(
            "create",
            create,
            "create a packaged model",
            [
                Arg(
                    "name",
                    help="The name of the packaged model. Must begin with a letter, but may "
                    "contain letters, numbers, and hyphen",
                ),
                Arg("--image", help="Docker container image servicing the packaged model"),
            ]
            + common_model_args,
        ),
        # dashboard command.
        Cmd(
            "dashboard",
            dashboard,
            "launch the packaged model dashboard",
            [
                Arg(
                    "name",
                    help=VERSIONED_MODEL_HELP_MSG,
                ),
            ],
        ),
        # Show command.
        Cmd(
            "show",
            show_model,
            "show a packaged model",
            [
                Arg(
                    "name",
                    help=VERSIONED_MODEL_HELP_MSG,
                ),
                Group(
                    Arg("--yaml", action="store_true", help="print as YAML", default=True),
                    Arg("--json", action="store_true", help="print as JSON"),
                ),
            ],
        ),
        # Update command
        Cmd(
            "update",
            update,
            "modify a packaged model",
            [
                Arg(
                    "modelname",
                    help=VERSIONED_MODEL_HELP_MSG,
                ),
                Arg(
                    "--name",
                    help="The new name of the packaged model. Must begin with a letter, but may "
                    "contain letters, numbers, and hyphen",
                ),
                Arg("--image", help="Docker container image servicing the packaged model"),
            ]
            + common_model_args,
        ),
        Cmd(
            "delete",
            delete_model,
            "delete a packaged model",
            [
                Arg(
                    "name",
                    help=VERSIONED_MODEL_HELP_MSG,
                ),
            ],
        ),
        Cmd(
            "token",
            auth_token,
            "get packaged model auth token",
            [
                Arg(
                    "name",
                    help=VERSIONED_MODEL_HELP_MSG,
                ),
            ],
        ),
        Cmd(
            "list-deployments",
            list_deployments,
            "list of deployment versions for a packaged model",
            [
                Arg(
                    "name",
                    help=VERSIONED_MODEL_HELP_MSG,
                ),
                Group(
                    Arg("--csv", action="store_true", help="print as CSV"),
                    Arg("--json", action="store_true", help="print as JSON"),
                    Arg("--yaml", action="store_true", help="print as YAML"),
                ),
            ],
        ),
    ],
)

args_description = [main_cmd]  # type: List[Any]



# coding: utf-8

"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictInt, StrictStr
from pydantic import Field
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class DeploymentModelVersion(BaseModel):
    """
    The DeploymentModelVersion provides model evolution data that track model deployment changes over time.
    """ # noqa: E501
    canary_traffic_percent: Optional[StrictInt] = Field(default=None, description="Percent traffic to pass to the model.", alias="canaryTrafficPercent")
    deployed: Optional[StrictStr] = Field(default=None, description="The deployment time. This is a read-only field and is automatically assigned on creation.")
    deployment_id: Optional[StrictStr] = Field(default=None, description="The deployment ID", alias="deploymentId")
    deployment_name: Optional[StrictStr] = Field(default=None, description="The deployment name", alias="deploymentName")
    model: Optional[StrictStr] = Field(default=None, description="The name of the model.")
    mdl_id: Optional[StrictStr] = Field(default=None, description="The ID of the model.", alias="modelId")
    mdl_version: Optional[StrictStr] = Field(default=None, description="The version of the model.", alias="modelVersion")
    native_app_name: Optional[StrictStr] = Field(default=None, description="The name of the Kubernetes application for the specific service version. Use this name to match the app value in Grafana/Prometheus to obtain logs and metrics for this deployed service version.", alias="nativeAppName")
    __properties: ClassVar[List[str]] = ["canaryTrafficPercent", "deployed", "deploymentId", "deploymentName", "model", "modelId", "modelVersion", "nativeAppName"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of DeploymentModelVersion from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of DeploymentModelVersion from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "canaryTrafficPercent": obj.get("canaryTrafficPercent"),
            "deployed": obj.get("deployed"),
            "deploymentId": obj.get("deploymentId"),
            "deploymentName": obj.get("deploymentName"),
            "model": obj.get("model"),
            "modelId": obj.get("modelId"),
            "modelVersion": obj.get("modelVersion"),
            "nativeAppName": obj.get("nativeAppName")
        })
        return _obj





# coding: utf-8

"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictInt, StrictStr
from pydantic import Field
from aiolirest.models.autoscaling import Autoscaling
from aiolirest.models.deployment_state import DeploymentState
from aiolirest.models.event_info import EventInfo
from aiolirest.models.security import Security
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class Deployment(BaseModel):
    """
    Deployment describes the deployment of a service.
    """ # noqa: E501
    arguments: Optional[List[StrictStr]] = Field(default=None, description="Arguments to be added to the service command line")
    auto_scaling: Optional[Autoscaling] = Field(default=None, alias="autoScaling")
    canary_traffic_percent: Optional[StrictInt] = Field(default=None, description="Percent traffic to pass to new model version", alias="canaryTrafficPercent")
    cluster_name: Optional[StrictStr] = Field(default=None, description="Name of the cluster.", alias="clusterName")
    environment: Optional[Dict[str, StrictStr]] = Field(default=None, description="Environment variables added to the service")
    goal_status: Optional[StrictStr] = Field(default='Ready', description="Specifies the intended status to be achieved by the deployment.  Supported values are: * `Ready` - The inference service will be deployed to enable inference calls. * `Paused` - The inference service will be stopped and no longer accept calls.  The default is `Ready`.", alias="goalStatus")
    id: Optional[StrictStr] = Field(default=None, description="The ID of the deployment. This is a read-only property and is automatically assigned on creation.")
    last_event: Optional[EventInfo] = Field(default=None, alias="lastEvent")
    model: StrictStr = Field(description="PackagedModel name or ID to be deployed.")
    modified_at: Optional[StrictStr] = Field(default=None, description="Date-time of last modification of the deployment. This is a read-only field and is automatically updated.", alias="modifiedAt")
    name: StrictStr = Field(description="The deployment name.  It must be a valid subdomain name and consist of lower case alphanumeric characters, '-' or '.', and must start and end with an alphanumeric character.")
    namespace: Optional[StrictStr] = Field(default=None, description="The Kubernetes namespace to be used for the deployment.")
    node_selectors: Optional[Dict[str, StrictStr]] = Field(default=None, description="Node selectors to be added to the service", alias="nodeSelectors")
    priority_class_name: Optional[StrictStr] = Field(default=None, description="Priority Class Name to be used for prioritization of deployments.", alias="priorityClassName")
    project: Optional[StrictStr] = Field(default=None, description="The ID of the containing project")
    secondary_state: Optional[DeploymentState] = Field(default=None, alias="secondaryState")
    security: Optional[Security] = None
    state: Optional[DeploymentState] = None
    status: Optional[StrictStr] = Field(default=None, description="Summary status of the deployed service. * `Deploying` - Service configuration is in progress. * `Failed` - The service configuration failed. * `Ready` - The service has been successfully configured and is serving. * `Updating` - A new service revision is being rolled-out. * `UpdateFailed` - The current service revision failed to rollout due to an error.   The prior version is still serving requests. * `Deleting` - The deployed service is being removed. * `Paused` - The deployed service has been stopped by the user or an external action. * `Unknown` - Unable to determined the status.  This is a read-only property. Must be one of the values: (Deploying,Ready,Updating,UpdateFailed,Failed,Deleting,Paused,Unknown).")
    __properties: ClassVar[List[str]] = ["arguments", "autoScaling", "canaryTrafficPercent", "clusterName", "environment", "goalStatus", "id", "lastEvent", "model", "modifiedAt", "name", "namespace", "nodeSelectors", "priorityClassName", "project", "secondaryState", "security", "state", "status"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of Deployment from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of auto_scaling
        if self.auto_scaling:
            _dict['autoScaling'] = self.auto_scaling.to_dict()
        # override the default output from pydantic by calling `to_dict()` of last_event
        if self.last_event:
            _dict['lastEvent'] = self.last_event.to_dict()
        # override the default output from pydantic by calling `to_dict()` of secondary_state
        if self.secondary_state:
            _dict['secondaryState'] = self.secondary_state.to_dict()
        # override the default output from pydantic by calling `to_dict()` of security
        if self.security:
            _dict['security'] = self.security.to_dict()
        # override the default output from pydantic by calling `to_dict()` of state
        if self.state:
            _dict['state'] = self.state.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of Deployment from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "arguments": obj.get("arguments"),
            "autoScaling": Autoscaling.from_dict(obj.get("autoScaling")) if obj.get("autoScaling") is not None else None,
            "canaryTrafficPercent": obj.get("canaryTrafficPercent"),
            "clusterName": obj.get("clusterName"),
            "environment": obj.get("environment"),
            "goalStatus": obj.get("goalStatus") if obj.get("goalStatus") is not None else 'Ready',
            "id": obj.get("id"),
            "lastEvent": EventInfo.from_dict(obj.get("lastEvent")) if obj.get("lastEvent") is not None else None,
            "model": obj.get("model"),
            "modifiedAt": obj.get("modifiedAt"),
            "name": obj.get("name"),
            "namespace": obj.get("namespace"),
            "nodeSelectors": obj.get("nodeSelectors"),
            "priorityClassName": obj.get("priorityClassName"),
            "project": obj.get("project"),
            "secondaryState": DeploymentState.from_dict(obj.get("secondaryState")) if obj.get("secondaryState") is not None else None,
            "security": Security.from_dict(obj.get("security")) if obj.get("security") is not None else None,
            "state": DeploymentState.from_dict(obj.get("state")) if obj.get("state") is not None else None,
            "status": obj.get("status")
        })
        return _obj





# coding: utf-8

"""
    HPE Machine Learning Inference Software (MLIS/Aioli)

    HPE MLIS is *Aioli* -- The AI On-line Inference Platform that enables easy deployment, tracking, and serving of your packaged models regardless of your preferred AI framework.

    The version of the OpenAPI document: 1.3.0-dev21
    Contact: community@determined-ai
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional
from pydantic import BaseModel, StrictInt, StrictStr
from pydantic import Field
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class ModelResponse(BaseModel):
    """
    Provides a set of values that are returned by querying for models
    """ # noqa: E501
    description: Optional[StrictStr] = Field(default=None, description="The description of desribing the model")
    display_name: Optional[StrictStr] = Field(default=None, description="The name displayed that is human readable", alias="displayName")
    image: Optional[StrictStr] = Field(default=None, description="The default container image to execute the model (if available).  For NIM models, this is used to identify the model.")
    latest_version_id_str: Optional[StrictStr] = Field(default=None, description="The latest version of the model", alias="latestVersionIdStr")
    latest_version_size_in_bytes: Optional[StrictInt] = Field(default=None, description="The number of bytes for the latest model (if available)", alias="latestVersionSizeInBytes")
    metadata: Optional[Dict[str, StrictStr]] = Field(default=None, description="Other attributes of the model as available")
    format: Optional[StrictStr] = Field(default=None, description="The format of the model", alias="modelFormat")
    name: Optional[StrictStr] = Field(default=None, description="The name used in model specification")
    url: Optional[StrictStr] = Field(default=None, description="The URL where the model is stored.")
    __properties: ClassVar[List[str]] = ["description", "displayName", "image", "latestVersionIdStr", "latestVersionSizeInBytes", "metadata", "modelFormat", "name", "url"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of ModelResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of ModelResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "description": obj.get("description"),
            "displayName": obj.get("displayName"),
            "image": obj.get("image"),
            "latestVersionIdStr": obj.get("latestVersionIdStr"),
            "latestVersionSizeInBytes": obj.get("latestVersionSizeInBytes"),
            "metadata": obj.get("metadata"),
            "modelFormat": obj.get("modelFormat"),
            "name": obj.get("name"),
            "url": obj.get("url")
        })
        return _obj





# © Copyright 2024 Hewlett Packard Enterprise Development LP

import argparse
import sys
from typing import Any, Dict

import termcolor
from packaging import version
from urllib3.exceptions import MaxRetryError, SSLError

import aioli
import aioli.cli
import aiolirest
from aioli import cli
from aioli.cli import render
from aioli.common.declarative_argparse import ArgsDescription, Cmd


def get_version(controller: str) -> Dict[str, Any]:
    client_info = {"version": aioli.__version__}

    controller_info = {"version": ""}

    with cli.setup_session_no_auth(controller) as session:
        api_instance = aiolirest.InformationApi(session)

        try:
            response = api_instance.version_get()
            controller_info["version"] = response
        except MaxRetryError as ex:
            # Most connection errors mean that the controller is unreachable, which this
            # function handles. An SSLError, however, means it was reachable but something
            # went wrong, so let that error propagate out.
            if ex.__cause__ and isinstance(ex.__cause__, SSLError):
                raise ex.__cause__
        except Exception:
            # Exceptions get a pass here so that the code in check_version can complete.
            pass

    return {
        "client": client_info,
        "controller": controller_info,
        "controller_address": controller,
    }


def check_version(controller: str) -> None:
    info = get_version(controller)

    controller_version = info["controller"]["version"]
    client_version = info["client"]["version"]

    if not controller_version:
        print(
            termcolor.colored(
                "Controller not found at {}. "
                "Hint: Remember to set the AIOLI_CONTROLLER environment variable "
                "to the correct controller IP and port, for example, "
                "export AIOLI_CONTROLLER=http://$AIOLI_IP:80. Or use the '-c' flag.".format(
                    controller
                ),
                "yellow",
            ),
            file=sys.stderr,
        )
    elif version.Version(client_version) < version.Version(controller_version):
        print(
            termcolor.colored(
                "CLI version {} is less than controller version {}. "
                "Consider upgrading the CLI.".format(client_version, controller_version),
                "yellow",
            ),
            file=sys.stderr,
        )
    elif version.Version(client_version) > version.Version(controller_version):
        print(
            termcolor.colored(
                "Controller version {} is less than CLI version {}. "
                "This CLI may utilize features not supported by an older "
                "controller. Install a compatible CLI with: "
                " pip install aioli-sdk=={}".format(
                    controller_version, client_version, controller_version
                ),
                "yellow",
            ),
            file=sys.stderr,
        )


def describe_version(parsed_args: argparse.Namespace) -> None:
    info = get_version(parsed_args.controller)
    print(render.format_object_as_yaml(info))


args_description: ArgsDescription = [
    Cmd("version", describe_version, "show version information", [])
]
